﻿using System;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;

using GHI.Processor;

namespace Cerberus_PWM
{
    class GetSettings
    {
        #region Class Variables...

        private bool debug = true;

        #endregion

        /// <summary>
        /// Get and Display Settings in the Debug Window.
        /// </summary>
        /// <param name="GPIOx_Base">STM32F4xx.GPIOA_BASE</param>
        /// <param name="TIMx_BASE">STM32F4xx.TIM1_BASE</param>
        public void GetRegisterSettings(UInt32 GPIOx_Base, UInt32 TIMx_BASE)
        {
            // Set the Clocks...
            ConfRCC();

            // Configure GPIO Port...
            ConfGPIO(GPIOx_Base);

            // Configure the Timer...
            ConfTimer(TIMx_BASE);
        }

        /// <summary>
        /// Enable RCC, Reset and clock control.
        /// </summary>
        private void ConfRCC()
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Reset and clock control for STM32F405xx/07xx and STM32F415xx/17xx(RCC)                               //
            //////////////////////////////////////////////////////////////////////////////////////////////////////////

            #region OFFSET's...

            UInt32 CR_OFFSET = 0x00;
            UInt32 PLLCFGR_OFFSET = 0x04;
            UInt32 CFGR_OFFSET = 0x08;
            UInt32 CIR_OFFSET = 0x0C;
            UInt32 AHB1RSTR_OFFSET = 0x10;
            UInt32 AHB2RSTR_OFFSET = 0x14;
            UInt32 AHB3RSTR_OFFSET = 0x18;
            UInt32 APB1RSTR_OFFSET = 0x20;
            UInt32 APB2RSTR_OFFSET = 0x24;
            UInt32 AHB1ENR_OFFSET = 0x30;
            UInt32 AHB2ENR_OFFSET = 0x34;
            UInt32 AHB3ENR_OFFSET = 0x38;
            UInt32 APB1ENR_OFFSET = 0x40;
            UInt32 APB2ENR_OFFSET = 0x44;
            UInt32 AHB1LPENR_OFFSET = 0x50;
            UInt32 AHB2LPENR_OFFSET = 0x54;
            UInt32 AHB3LPENR_OFFSET = 0x58;
            UInt32 APB1LPENR_OFFSET = 0x60;
            UInt32 APB2LPENR_OFFSET = 0x64;
            UInt32 BDCR_OFFSET = 0x70;
            UInt32 CSR_OFFSET = 0x74;
            UInt32 SSCGR_OFFSET = 0x80;
            UInt32 PLLI2SCFGR_OFFSET = 0x84;

            #endregion

            #region Registers...

            // RCC clock control register (RCC_CR)
            // Address offset: 0x00
            // Reset value: 0x0000 XX83 where X is undefined.
            // Access: no wait state, word, half-word and byte access
            Register CR = new Register(STM32F4xx.RCC_BASE + CR_OFFSET);

            // RCC PLL configuration register (RCC_PLLCFGR)
            // Address offset: 0x04
            // Reset value: 0x2400 3010
            // Access: no wait state, word, half-word and byte access.
            // This register is used to configure the PLL clock outputs according to the formulas:
            // • f(VCO clock) = f(PLL clock input) × (PLLN / PLLM)
            // • f(PLL general clock output) = f(VCO clock) / PLLP
            // • f(USB OTG FS, SDIO, RNG clock output) = f(VCO clock) / PLLQ
            Register PLLCFGR = new Register(STM32F4xx.RCC_BASE + PLLCFGR_OFFSET);

            // RCC clock configuration register (RCC_CFGR)
            // Address offset: 0x08
            // Reset value: 0x0000 0000
            // Access: 0 ≤ wait state ≤ 2, word, half-word and byte access
            // 1 or 2 wait states inserted only if the access occurs during a clock source switch.
            Register CFGR = new Register(STM32F4xx.RCC_BASE + CFGR_OFFSET);

            // RCC clock interrupt register (RCC_CIR)
            // Address offset: 0x0C
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access
            Register CIR = new Register(STM32F4xx.RCC_BASE + CIR_OFFSET);

            // RCC AHB1 peripheral reset register (RCC_AHB1RSTR)
            // Address offset: 0x10
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register AHB1RSTR = new Register(STM32F4xx.RCC_BASE + AHB1RSTR_OFFSET);

            // RCC AHB2 peripheral reset register (RCC_AHB2RSTR)
            // Address offset: 0x14
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access
            Register AHB2RSTR = new Register(STM32F4xx.RCC_BASE + AHB2RSTR_OFFSET);

            // RCC AHB3 peripheral reset register (RCC_AHB3RSTR)
            // Address offset: 0x18
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register AHB3RSTR = new Register(STM32F4xx.RCC_BASE + AHB3RSTR_OFFSET);

            // RCC APB1 peripheral reset register (RCC_APB1RSTR)
            // Address offset: 0x20
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register APB1RSTR = new Register(STM32F4xx.RCC_BASE + APB1RSTR_OFFSET);

            // RCC APB2 peripheral reset register (RCC_APB2RSTR)
            // Address offset: 0x24
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register APB2RSTR = new Register(STM32F4xx.RCC_BASE + APB2RSTR_OFFSET);

            // RCC AHB1 peripheral clock register (RCC_AHB1ENR)
            // Address offset: 0x30
            // Reset value: 0x0010 0000
            // Access: no wait state, word, half-word and byte access.
            Register AHB1ENR = new Register(STM32F4xx.RCC_BASE + AHB1ENR_OFFSET);

            // RCC AHB2 peripheral clock enable register (RCC_AHB2ENR)
            // Address offset: 0x34
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register AHB2ENR = new Register(STM32F4xx.RCC_BASE + AHB2ENR_OFFSET);

            // RCC AHB3 peripheral clock enable register (RCC_AHB3ENR)
            // Address offset: 0x38
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register AHB3ENR = new Register(STM32F4xx.RCC_BASE + AHB3ENR_OFFSET);

            // RCC APB1 peripheral clock enable register (RCC_APB1ENR)
            // Address offset: 0x40
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            // See: P.181
            Register APB1ENR = new Register(STM32F4xx.RCC_BASE + APB1ENR_OFFSET);

            // RCC APB2 peripheral clock enable register (RCC_APB2ENR)
            // Address offset: 0x44
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            // See: P.185
            Register APB2ENR = new Register(STM32F4xx.RCC_BASE + APB2ENR_OFFSET);

            // RCC AHB1 peripheral clock enable in low power mode register
            // (RCC_AHB1LPENR)
            // Address offset: 0x50
            // Reset value: 0x7E67 91FF
            // Access: no wait state, word, half-word and byte access.
            Register AHB1LPENR = new Register(STM32F4xx.RCC_BASE + AHB1LPENR_OFFSET);

            // RCC AHB2 peripheral clock enable in low power mode register
            // (RCC_AHB2LPENR)
            // Address offset: 0x54
            // Reset value: 0x0000 00F1
            // Access: no wait state, word, half-word and byte access.
            Register AHB2LPENR = new Register(STM32F4xx.RCC_BASE + AHB2LPENR_OFFSET);

            // RCC AHB3 peripheral clock enable in low power mode register
            // (RCC_AHB3LPENR)
            // Address offset: 0x58
            // Reset value: 0x0000 0001
            // Access: no wait state, word, half-word and byte access.
            Register AHB3LPENR = new Register(STM32F4xx.RCC_BASE + AHB3LPENR_OFFSET);

            // RCC APB1 peripheral clock enable in low power mode register
            // (RCC_APB1LPENR)
            // Address offset: 0x60
            // Reset value: 0x36FE C9FF
            // Access: no wait state, word, half-word and byte access.
            Register APB1LPENR = new Register(STM32F4xx.RCC_BASE + APB1LPENR_OFFSET);

            // RCC APB2 peripheral clock enabled in low power mode register
            // (RCC_APB2LPENR)
            // Address offset: 0x64
            // Reset value: 0x0007 5F33
            // Access: no wait state, word, half-word and byte access.
            Register APB2LPENR = new Register(STM32F4xx.RCC_BASE + APB2LPENR_OFFSET);

            // RCC Backup domain control register (RCC_BDCR)
            // Address offset: 0x70
            // Reset value: 0x0000 0000, reset by Backup domain reset.
            // Access: 0 ≤ wait state ≤ 3, word, half-word and byte access
            // Wait states are inserted in case of successive accesses to this register.
            Register BDCR = new Register(STM32F4xx.RCC_BASE + BDCR_OFFSET);

            // RCC clock control & status register (RCC_CSR)
            // Address offset: 0x74
            // Reset value: 0x0E00 0000, reset by system reset, except reset flags by power reset only.
            // Access: 0 ≤ wait state ≤ 3, word, half-word and byte access
            // Wait states are inserted in case of successive accesses to this register.
            Register CSR = new Register(STM32F4xx.RCC_BASE + CSR_OFFSET);

            // RCC spread spectrum clock generation register (RCC_SSCGR)
            // Address offset: 0x80
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            // The spread spectrum clock generation is available only for the main PLL.
            // The RCC_SSCGR register must be written either before the main PLL is enabled or after
            // the main PLL disabled.
            // Note: For full details about PLL spread spectrum clock generation (SSCG) characteristics, refer to
            // the “Electrical characteristics” section in your device datasheet.
            Register SSCGR = new Register(STM32F4xx.RCC_BASE + SSCGR_OFFSET);

            // RCC PLLI2S configuration register (RCC_PLLI2SCFGR)
            // Address offset: 0x84
            // Reset value: 0x2000 3000
            // Access: no wait state, word, half-word and byte access.
            // This register is used to configure the PLLI2S clock outputs according to the formulas:
            // f(VCO clock) = f(PLLI2S clock input) × (PLLI2SN / PLLM)
            // f(PLL I2S clock output) = f(VCO clock) / PLLI2SR
            Register PLLI2SCFGR = new Register(STM32F4xx.RCC_BASE + PLLI2SCFGR_OFFSET);

            #endregion

            #region Print Values...

            if (debug)
            {
                // Print to Debug window current config...
                Debug.Print("RCC **************");

                // Print Clock Configuration Register...
                Print("CFGR", CFGR);

                // peripheral clock register's...
                Print("AHB1ENR", AHB1ENR);
                Print("AHB2ENR", AHB2ENR);
                Print("AHB3ENR", AHB3ENR);

                // peripheral clock enable register's...
                Print("APB1ENR", APB1ENR);
                Print("APB2ENR", APB2ENR);
            }

            #endregion
        }

        /// <summary>
        /// Configure GPIO Ports.
        /// Each of the 15 GPIOx Pins on each GPIO Port (GPIOA through I) is mapped to the MODER/AF/OSPEEDR/PUPDR and so on.
        /// This allows the control of each Pin, on each port to have Multiplexed Functions.
        /// Example: For GPIOC, MODER6 is mapped to Pin PA6, AF3 set on this Port in the AFRL (because the Pin is less than 7) sets TIM8 CH1.
        /// See: STM32F405VG User Manual P.278 and STM32F405VG P.62 for mappings.
        /// </summary>
        /// <param name="GPIOx_BASE">The Register BASE Address: 'STM32F4xx.GPIOC_BASE'</param>
        private void ConfGPIO(UInt32 GPIOx_BASE)
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            // General-purpose I/O (GPIO) for STM32F405xx/07xx and STM32F415xx/17xx(RCC)                            //
            //////////////////////////////////////////////////////////////////////////////////////////////////////////

            #region OFFSET's...

            UInt32 MODER_OFFSET = 0x00;
            UInt32 OTYPER_OFFSET = 0x04;
            UInt32 OSPEEDR_OFFSET = 0x08;
            UInt32 PUPDR_OFFSET = 0x0C;
            UInt32 IDR_OFFSET = 0x10;
            UInt32 ODR_OFFSET = 0x14;
            UInt32 BSRR_OFFSET = 0x18;
            // UInt32 BSRRH_OFFSET = 0x1A; <--- ?
            UInt32 LCKR_OFFSET = 0x1C;

            #endregion

            #region Register's...

            // GPIO port mode register (GPIOx_MODER) (x = A..I/J/K)
            // Address offset: 0x00
            // Reset values:
            // • 0xA800 0000 for port A
            // • 0x0000 0280 for port B
            // • 0x0000 0000 for other ports
            Register MODER = new Register(GPIOx_BASE + MODER_OFFSET);

            // GPIO port output type register (GPIOx_OTYPER)
            // (x = A..I/J/K)
            // Address offset: 0x04
            // Reset value: 0x0000 0000
            Register OTYPER = new Register(GPIOx_BASE + OTYPER_OFFSET);

            // GPIO port output speed register (GPIOx_OSPEEDR)
            // (x = A..I/J/K)
            // Address offset: 0x08
            // Reset values:
            // • 0x0000 00C0 for port B
            // • 0x0000 0000 for other ports
            Register OSPEEDR = new Register(GPIOx_BASE + OSPEEDR_OFFSET);

            // GPIO port pull-up/pull-down register (GPIOx_PUPDR)
            // (x = A..I/J/K)
            // Address offset: 0x0C
            // Reset values:
            // • 0x6400 0000 for port A
            // • 0x0000 0100 for port B
            // • 0x0000 0000 for other ports
            Register PUPDR = new Register(GPIOx_BASE + PUPDR_OFFSET);

            // GPIO port input data register (GPIOx_IDR) (x = A..I/J/K)
            // Address offset: 0x10
            // Reset value: 0x0000 XXXX (where X means undefined)
            Register IDR = new Register(GPIOx_BASE + IDR_OFFSET);

            // GPIO port output data register (GPIOx_ODR) (x = A..I/J/K)
            // Address offset: 0x14
            // Reset value: 0x0000 0000
            Register ODR = new Register(GPIOx_BASE + ODR_OFFSET);

            // GPIO port bit set/reset register (GPIOx_BSRR) (x = A..I/J/K)
            // Address offset: 0x18
            // Reset value: 0x0000 0000
            Register BSRR = new Register(GPIOx_BASE + BSRR_OFFSET);

            // GPIO port configuration lock register (GPIOx_LCKR)
            // (x = A..I/J/K)
            // This register is used to lock the configuration of the port bits when a correct write sequence
            // is applied to bit 16 (LCKK). The value of bits [15:0] is used to lock the configuration of the
            // GPIO. During the write sequence, the value of LCKR[15:0] must not change. When the
            // LOCK sequence has been applied on a port bit, the value of this port bit can no longer be
            // modified until the next reset.
            // Note: A specific write sequence is used to write to the GPIOx_LCKR register. Only word access
            // (32-bit long) is allowed during this write sequence.
            // Each lock bit freezes a specific configuration register (control and alternate function
            // registers).
            // Address offset: 0x1C
            // Reset value: 0x0000 0000
            // Access: 32-bit word only, read/write register
            Register LCKR = new Register(GPIOx_BASE + LCKR_OFFSET);

            // Alternate Functions...
            // GPIO alternate function low register (GPIOx_AFRL) (x = A..I/J/K) - AFR[0]
            // Address offset: 0x20
            // Reset value: 0x0000 0000
            // GPIO alternate function high register (GPIOx_AFRH) - AFR[1]
            // (x = A..I/J/K)
            // Address offset: 0x24
            // Reset value: 0x0000 0000
            Register[] AFR = new Register[] { new Register(GPIOx_BASE + 0x20), new Register(GPIOx_BASE + 0x24) };

            #endregion

            #region Print Values...

            if (debug)
            {
                // Print to Debug window current config...
                Debug.Print("GPIO *************");

                Print("AFR[0]", AFR[0]);
                Print("AFR[1]", AFR[1]);
                Print("MODER", MODER);
                Print("OTYPER", OTYPER);
                Print("OSPEEDR", OSPEEDR);
                Print("PUPDR", PUPDR);
            }

            #endregion
        }

        /// <summary>
        /// Timer Configuration.
        /// </summary>
        /// <param name="TIMx_BASE">The Register BASE Address: 'STM32F4xx.TIM8_BASE'</param>
        private void ConfTimer(UInt32 TIMx_BASE)
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            // General-purpose I/O (GPIO) for STM32F405xx/07xx and STM32F415xx/17xx(RCC)                            //
            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Good Resources:
            // Table 42. DMA1 request mapping - See: P.304
            // Table 43. DMA2 request mapping - See: P.305
            // Advanced-control timers (TIM1&TIM8) - See: P.507

            #region OFFSET's...

            UInt32 CR1_OFFSET = 0x00;
            UInt32 CR2_OFFSET = 0x04;
            UInt32 SMCR_OFFSET = 0x08;
            UInt32 DIER_OFFSET = 0x0C;
            UInt32 SR_OFFSET = 0x10;
            UInt32 EGR_OFFSET = 0x14;
            UInt32 CCMR1_OFFSET = 0x18;
            UInt32 CCMR2_OFFSET = 0x1C;
            UInt32 CCER_OFFSET = 0x20;
            UInt32 CNT_OFFSET = 0x24;
            UInt32 PSC_OFFSET = 0x28;
            UInt32 ARR_OFFSET = 0x2C;
            UInt32 RCR_OFFSET = 0x30;
            UInt32 CCR1_OFFSET = 0x34;
            UInt32 CCR2_OFFSET = 0x38;
            UInt32 CCR3_OFFSET = 0x3C;
            UInt32 CCR4_OFFSET = 0x40;
            UInt32 BDTR_OFFSET = 0x44;
            UInt32 DCR_OFFSET = 0x48;
            UInt32 DMAR_OFFSET = 0x4C;
            UInt32 OR_OFFSET = 0x50;

            #endregion

            #region Registers...

            // TIM1&TIM8 control register 1 (TIMx_CR1)
            // Address offset: 0x00
            // Reset value: 0x0000
            Register CR1 = new Register(TIMx_BASE + CR1_OFFSET);

            // TIM1&TIM8 control register 2 (TIMx_CR2)
            // Address offset: 0x04
            // Reset value: 0x0000
            Register CR2 = new Register(TIMx_BASE + CR2_OFFSET);

            // TIM1&TIM8 slave mode control register (TIMx_SMCR)
            // Address offset: 0x08
            // Reset value: 0x0000
            Register SMCR = new Register(TIMx_BASE + SMCR_OFFSET);

            // TIM1&TIM8 DMA/interrupt enable register (TIMx_DIER)
            // Address offset: 0x0C
            // Reset value: 0x0000
            Register DIER = new Register(TIMx_BASE + DIER_OFFSET);

            // TIM1&TIM8 status register (TIMx_SR)
            // Address offset: 0x10
            // Reset value: 0x0000
            Register SR = new Register(TIMx_BASE + SR_OFFSET);

            // TIM1&TIM8 event generation register (TIMx_EGR)
            // Address offset: 0x14
            // Reset value: 0x0000
            Register EGR = new Register(TIMx_BASE + EGR_OFFSET);

            // TIM1&TIM8 capture/compare mode register 1 (TIMx_CCMR1)
            // Address offset: 0x18
            // Reset value: 0x0000
            Register CCMR1 = new Register(TIMx_BASE + CCMR1_OFFSET);

            // TIM1&TIM8 capture/compare mode register 2 (TIMx_CCMR2)
            // Address offset: 0x1C
            // Reset value: 0x0000
            // Refer to the above CCMR1 register description.
            Register CCMR2 = new Register(TIMx_BASE + CCMR2_OFFSET);

            // TIM1&TIM8 capture/compare enable register (TIMx_CCER)
            // Address offset: 0x20
            // Reset value: 0x0000
            Register CCER = new Register(TIMx_BASE + CCER_OFFSET);

            // TIM1&TIM8 counter (TIMx_CNT)
            // Address offset: 0x24
            // Reset value: 0x0000
            Register CNT = new Register(TIMx_BASE + CNT_OFFSET);

            // TIM1&TIM8 prescaler (TIMx_PSC)
            // Address offset: 0x28
            // Reset value: 0x0000
            Register PSC = new Register(TIMx_BASE + PSC_OFFSET);

            // TIM1&TIM8 auto-reload register (TIMx_ARR)
            // Address offset: 0x2C
            // Reset value: 0x0000
            Register ARR = new Register(TIMx_BASE + ARR_OFFSET);

            // TIM1&TIM8 repetition counter register (TIMx_RCR)
            // Address offset: 0x30
            // Reset value: 0x0000
            Register RCR = new Register(TIMx_BASE + RCR_OFFSET);

            // TIM1&TIM8 capture/compare register 1 (TIMx_CCR1)
            // Address offset: 0x34
            // Reset value: 0x0000
            Register CCR1 = new Register(TIMx_BASE + CCR1_OFFSET);

            // TIM1&TIM8 capture/compare register 2 (TIMx_CCR2)
            // Address offset: 0x38
            // Reset value: 0x0000
            Register CCR2 = new Register(TIMx_BASE + CCR2_OFFSET);

            // TIM1&TIM8 capture/compare register 3 (TIMx_CCR3)
            // Address offset: 0x3C
            // Reset value: 0x0000
            Register CCR3 = new Register(TIMx_BASE + CCR3_OFFSET);

            // TIM1&TIM8 capture/compare register 4 (TIMx_CCR4)
            // Address offset: 0x40
            // Reset value: 0x0000
            Register CCR4 = new Register(TIMx_BASE + CCR4_OFFSET);

            // TIM1&TIM8 break and dead-time register (TIMx_BDTR)
            // Address offset: 0x44
            // Reset value: 0x0000
            Register BDTR = new Register(TIMx_BASE + BDTR_OFFSET);

            // TIM1&TIM8 DMA control register (TIMx_DCR)
            // Address offset: 0x48
            // Reset value: 0x0000
            Register DCR = new Register(TIMx_BASE + DCR_OFFSET);

            // TIM1&TIM8 DMA address for full transfer (TIMx_DMAR)
            // Address offset: 0x4C
            // Reset value: 0x0000
            Register DMAR = new Register(TIMx_BASE + DMAR_OFFSET);

            #endregion

            #region Print Values...

            if (debug)
            {
                // Print to Debug window current config...
                Debug.Print("Timer ************");
                Print("CR1", CR1);
                Print("CR2", CR2);
                Print("SMCR", SMCR);
                Print("DIER", DIER);
                Print("SR", SR);
                Print("EGR", EGR);
                Print("CCMR1", CCMR1);
                Print("CCMR2", CCMR2);
                Print("CCER", CCER);
                Print("CNT", CNT);
                Print("PSC", PSC);
                Print("ARR", ARR);
                Print("RCR", RCR);
                Print("CCR1", CCR1);
                Print("CCR2", CCR2);
                Print("CCR3", CCR3);
                Print("CCR4", CCR4);
                Print("BDTR", BDTR);
                Print("DCR", DCR);
                Print("DMAR", DMAR);
                Debug.Print("******************");
            }

            #endregion
        }

        /// <summary>
        /// Debug.Print the Register Name and the current Address...
        /// </summary>
        /// <param name="Name">String Register Display Name: 'CNT' </param>
        /// <param name="Address">The Register to be read: 'CNT'</param>
        private void Print(String Name, Register Register)
        {
            Debug.Print(Name + ": " + Register.Value);
        }

    }
}
