﻿using System;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;

using System.Collections;

using GHI.OSHW.Hardware.LowLevel;

namespace Cerberus_PWM
{
    public class Timer
    {
        #region Fields...

        // For Debugging Purposes... 
        private static bool debug = false;

        private static uint m_period;

        private static ScaleFactor m_scale;

        private static Channel m_channel;

        private static CPU.Clock m_busclock;

        private static CPU.Pin m_pin;

        private static uint m_prescaler;

        private static uint m_duration;

        private static bool m_invert;

        private static uint ARRValue;

        private static uint PSCValue;

        private static uint CCRxValue;

        private UInt32 GPIOx_BASE = 0;

        private UInt32 TIMx_BASE = 0;

        #endregion

        #region Properties...

        public static CPU.Clock ClockSpeed
        {
            get { return m_busclock; }
            set { m_busclock = value; }
        }

        public CPU.Pin PWMPin
        {
            get { return m_pin; }
            set { m_pin = value; }
        }

        public ScaleFactor Scale
        {
            get
            {
                return m_scale;
            }
            set
            {
                m_scale = value;
            }
        }

        public Channel PWMChannel
        {
            get
            {
                return m_channel;
            }
            set
            {
                m_channel = value;
            }
        }

        public double Frequency
        {
            get
            {
                return FrequencyFromPeriod((double)m_period, m_scale);
            }
            set
            {
                m_period = PeriodFromFrequency(value, out m_scale);
            }
        }

        public uint Prescaler
        {
            get { return m_prescaler; }
            set { m_prescaler = value; }
        }

        public uint Period
        {
            get
            {
                return m_period;
            }
            set
            {
                m_period = value;
            }
        }

        public double DutyCycle
        {
            get
            {
                return DutyCycleFromDurationAndPeriod((double)m_period, (double)m_duration);
            }
            set
            {
                m_duration = DurationFromDutyCycleAndPeriod(value, (double)m_period);
            }
        }

        public uint Duration
        {
            get
            {
                return m_duration;
            }
            set
            {
                m_duration = value;
            }
        }

        public bool InvertPin
        {
            get { return m_invert; }
            set { m_invert = value; }
        }

        #endregion

        public enum ScaleFactor : uint
        {
            Microseconds = 0xf4240,
            Milliseconds = 0x3e8,
            Nanoseconds = 0x3b9aca00
        }

        /// <summary>
        /// Configure the Timer CCRx Channel. Each Timer can have up to 4 Channels.
        /// </summary>
        public enum Channel
        {
            CH1,
            CH2,
            CH3,
            CH4
        }

        /// <summary>
        /// Set the PWM with the desired Params...
        /// </summary>
        /// <param name="Frequency">Frequency of the PWM</param>
        /// <param name="DutyCycle">Double - 0.0 - 1 - Controls the Duty Cycle</param>
        /// <param name="InvertPin">Invert Chanel</param>
        public void SetPWM(CPU.Pin pin, double frequency_Hz, double dutyCycle, bool InvertPin)
        {
            try
            {
                // Configure current Pin...
                SetPinPort(pin);
            }
            catch
            {
                Debug.Print("Error setting the Pin Info in SetPWM...");
            }

            // Configure Values...
            this.Frequency = frequency_Hz;
            this.Period = PeriodFromFrequency(frequency_Hz, out m_scale);
            this.Duration = DurationFromDutyCycleAndPeriod(dutyCycle, (double)m_period);

            // Set Pin Invert...
            this.InvertPin = InvertPin;

            // Commit the Values to HAL...
            Commit();
        }

        /// <summary>
        /// Gets a calculated Value for the PSC (Prescaler) Register.
        /// </summary>
        /// <param name="period">The Period, the ARR Value that is written into the ARR Register.</param>
        /// <param name="frequency">Frequency, the Frequency that was assigned to the SetPWM Method.</param>
        /// <returns>UInt value for the Prescaler.</returns>
        private static uint PrescalerFromFrequencyAndClock(double period, double frequency)
        {
            return (uint)((((uint)ClockSpeed / ((period) / (1 / frequency))) + 0.5) - 1);
        }

        // Microsoft.SPOT.Hardware.PWM Method...
        private static uint PeriodFromFrequency(double f, out ScaleFactor scale)
        {
            if (f >= 1000.0)
            {
                scale = ScaleFactor.Nanoseconds;
                return (uint)((1000000000.0 / f) + 0.5);
            }
            if (f >= 1.0)
            {
                scale = ScaleFactor.Microseconds;
                return (uint)((1000000.0 / f) + 0.5);
            }
            scale = ScaleFactor.Milliseconds;
            return (uint)((1000.0 / f) + 0.5);
        }

        // Microsoft.SPOT.Hardware.PWM Method...
        private static double FrequencyFromPeriod(double period, ScaleFactor scale)
        {
            return (((double)scale) / period);
        }

        // Microsoft.SPOT.Hardware.PWM Method...
        private static uint DurationFromDutyCycleAndPeriod(double dutyCycle, double period)
        {
            if (period <= 0.0)
            {
                throw new ArgumentException();
            }
            if (dutyCycle < 0.0)
            {
                return 0;
            }
            if (dutyCycle > 1.0)
            {
                return 1;
            }
            return (uint)(dutyCycle * period);
        }

        // Microsoft.SPOT.Hardware.PWM Method...
        private static double DutyCycleFromDurationAndPeriod(double period, double duration)
        {
            if (period <= 0.0)
            {
                throw new ArgumentException();
            }
            if (duration < 0.0)
            {
                return 0.0;
            }
            if (duration > period)
            {
                return 1.0;
            }
            return (duration / period);
        }

        private void SetPinPort(CPU.Pin pin)
        {
            // Set the Pin...
            this.PWMPin = pin;

            /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Channels refer to the CCRx Registers. Some Timers can have 4 Channels per Timer. CCR1, CCR2, CCR3, and CCR4...          //
            // See: P.8 - AN4013 - http://www.st.com/web/en/resource/technical/document/application_note/DM00042534.pdf                //
            /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

            // Timer 8 CH1...
            if (pin == CPU.Pin.Pin7_X3)
            {
                PWMChannel = Channel.CH1;
                this.GPIOx_BASE = STM32F4xx.GPIOC_BASE;
                this.TIMx_BASE = STM32F4xx.TIM8_BASE;
            }
            // Timer 14 CH1...
            else if (pin == CPU.Pin.Pin8_X3)
            {
                PWMChannel = Channel.CH1;
                this.GPIOx_BASE = STM32F4xx.GPIOA_BASE;
                this.TIMx_BASE = STM32F4xx.TIM14_BASE;
            }
            // Timer 8 CH2...
            else if (pin == CPU.Pin.Pin9_X3)
            {
                PWMChannel = Channel.CH2;
                this.GPIOx_BASE = STM32F4xx.GPIOC_BASE;
                this.TIMx_BASE = STM32F4xx.TIM8_BASE;
            }
            // Timer 1 CH1...
            else if (pin == CPU.Pin.Pin7_X4)
            {
                PWMChannel = Channel.CH1;
                this.GPIOx_BASE = STM32F4xx.GPIOA_BASE;
                this.TIMx_BASE = STM32F4xx.TIM1_BASE;
            }
            // Timer 3 CH3...
            else if (pin == CPU.Pin.Pin8_X4)
            {
                PWMChannel = Channel.CH3;
                this.GPIOx_BASE = STM32F4xx.GPIOB_BASE;
                this.TIMx_BASE = STM32F4xx.TIM3_BASE;
            }
            // Timer 3 CH4...
            else if (pin == CPU.Pin.Pin9_X4)
            {
                PWMChannel = Channel.CH4;
                this.GPIOx_BASE = STM32F4xx.GPIOB_BASE;
                this.TIMx_BASE = STM32F4xx.TIM3_BASE;

                Debug.Print("Setting GPIO: " + this.GPIOx_BASE);
            }
        }

        /// <summary>
        /// Prepare and Commit the Values to the STM32F4xx Registers...
        /// </summary>
        /// <returns>true if success, false on failure.</returns>
        private bool Commit()
        {
            // Set the Bus Speed (APB1 or APB2), all timers are APB1 except TIM1 and TIM8...
            if (PWMPin == CPU.Pin.Pin7_X3 || PWMPin == CPU.Pin.Pin9_X3 || PWMPin == CPU.Pin.Pin7_X4)
            {
                ClockSpeed = CPU.Clock.APB2;
            }
            else
            {
                ClockSpeed = CPU.Clock.APB1;
            }

            #region Fields...

            uint ONE_MHZ = 1000000;

            uint prescaler = 0;
            uint period = m_period;
            uint duration = m_duration;
            uint scale = ((uint)m_scale);

            // Scale in MHz... 
            uint scaleMHz = scale / ONE_MHZ;

            uint clk = ((uint)ClockSpeed / ONE_MHZ);

            if (duration > period) duration = period;

            #endregion

            #region Calculate Variables (This is a C# implimentation of GHI Source ("PWM_ApplyConfiguration"))...

            prescaler = clk / scale;

            if (prescaler == 0)
            {
                if (period > (0xFFFFFFFF / clk))
                {
                    // Avoid Overflow... 
                    prescaler = clk;
                    period /= scaleMHz;
                    duration /= scaleMHz;
                }
                else
                {
                    prescaler = 1;
                    period = period * clk / scaleMHz;
                    duration = duration * clk / scaleMHz;
                }
            }
            else
            {
                while (prescaler > 0x10000)
                {
                    // Prescaler too Large... 
                    if (period >= 0x80000000) break;
                    prescaler >>= 1;
                    period <<= 1;
                    duration <<= 1;
                }
            }

            bool TimerIs16Bit = true;

            // Were not using Timer 2 of Timer 5...
            if (TimerIs16Bit)
            {
                // All Timers are 16 bit Timer's except 2 and 5... 
                while (period >= 0x10000)
                {
                    // period too large 
                    if (prescaler > 0x8000) break;
                    prescaler <<= 1;
                    period >>= 1;
                    duration >>= 1;
                }
            }

            #endregion

            ARRValue = (period - 1);
            PSCValue = PrescalerFromFrequencyAndClock(period, this.Frequency);
            CCRxValue = duration;

            try
            {
                // Enable the Clocks...
                ConfRCC();

                // Configure GPIO Port...
                ConfGPIO();

                // Configure the Timer...
                ConfTimer();

                return true;
            }
            catch
            {
                return false;
            }
        }

        #region Commit Values to the HAL...

        /// <summary>
        /// Enable RCC, Reset and clock control.
        /// </summary>
        private void ConfRCC()
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Reset and clock control for STM32F405xx/07xx and STM32F415xx/17xx(RCC)                               //
            //////////////////////////////////////////////////////////////////////////////////////////////////////////

            #region OFFSET's...

            UInt32 CR_OFFSET = 0x00;
            UInt32 PLLCFGR_OFFSET = 0x04;
            UInt32 CFGR_OFFSET = 0x08;
            UInt32 CIR_OFFSET = 0x0C;
            UInt32 AHB1RSTR_OFFSET = 0x10;
            UInt32 AHB2RSTR_OFFSET = 0x14;
            UInt32 AHB3RSTR_OFFSET = 0x18;
            UInt32 APB1RSTR_OFFSET = 0x20;
            UInt32 APB2RSTR_OFFSET = 0x24;
            UInt32 AHB1ENR_OFFSET = 0x30;
            UInt32 AHB2ENR_OFFSET = 0x34;
            UInt32 AHB3ENR_OFFSET = 0x38;
            UInt32 APB1ENR_OFFSET = 0x40;
            UInt32 APB2ENR_OFFSET = 0x44;
            UInt32 AHB1LPENR_OFFSET = 0x50;
            UInt32 AHB2LPENR_OFFSET = 0x54;
            UInt32 AHB3LPENR_OFFSET = 0x58;
            UInt32 APB1LPENR_OFFSET = 0x60;
            UInt32 APB2LPENR_OFFSET = 0x64;
            UInt32 BDCR_OFFSET = 0x70;
            UInt32 CSR_OFFSET = 0x74;
            UInt32 SSCGR_OFFSET = 0x80;
            UInt32 PLLI2SCFGR_OFFSET = 0x84;

            #endregion

            #region Registers...

            // RCC clock control register (RCC_CR)
            // Address offset: 0x00
            // Reset value: 0x0000 XX83 where X is undefined.
            // Access: no wait state, word, half-word and byte access
            Register CR = new Register(STM32F4xx.RCC_BASE + CR_OFFSET);

            // RCC PLL configuration register (RCC_PLLCFGR)
            // Address offset: 0x04
            // Reset value: 0x2400 3010
            // Access: no wait state, word, half-word and byte access.
            // This register is used to configure the PLL clock outputs according to the formulas:
            // • f(VCO clock) = f(PLL clock input) × (PLLN / PLLM)
            // • f(PLL general clock output) = f(VCO clock) / PLLP
            // • f(USB OTG FS, SDIO, RNG clock output) = f(VCO clock) / PLLQ
            Register PLLCFGR = new Register(STM32F4xx.RCC_BASE + PLLCFGR_OFFSET);

            // RCC clock configuration register (RCC_CFGR)
            // Address offset: 0x08
            // Reset value: 0x0000 0000
            // Access: 0 ≤ wait state ≤ 2, word, half-word and byte access
            // 1 or 2 wait states inserted only if the access occurs during a clock source switch.
            Register CFGR = new Register(STM32F4xx.RCC_BASE + CFGR_OFFSET);

            // RCC clock interrupt register (RCC_CIR)
            // Address offset: 0x0C
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access
            Register CIR = new Register(STM32F4xx.RCC_BASE + CIR_OFFSET);

            // RCC AHB1 peripheral reset register (RCC_AHB1RSTR)
            // Address offset: 0x10
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register AHB1RSTR = new Register(STM32F4xx.RCC_BASE + AHB1RSTR_OFFSET);

            // RCC AHB2 peripheral reset register (RCC_AHB2RSTR)
            // Address offset: 0x14
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access
            Register AHB2RSTR = new Register(STM32F4xx.RCC_BASE + AHB2RSTR_OFFSET);

            // RCC AHB3 peripheral reset register (RCC_AHB3RSTR)
            // Address offset: 0x18
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register AHB3RSTR = new Register(STM32F4xx.RCC_BASE + AHB3RSTR_OFFSET);

            // RCC APB1 peripheral reset register (RCC_APB1RSTR)
            // Address offset: 0x20
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register APB1RSTR = new Register(STM32F4xx.RCC_BASE + APB1RSTR_OFFSET);

            // RCC APB2 peripheral reset register (RCC_APB2RSTR)
            // Address offset: 0x24
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register APB2RSTR = new Register(STM32F4xx.RCC_BASE + APB2RSTR_OFFSET);

            // RCC AHB1 peripheral clock register (RCC_AHB1ENR)
            // Address offset: 0x30
            // Reset value: 0x0010 0000
            // Access: no wait state, word, half-word and byte access.
            Register AHB1ENR = new Register(STM32F4xx.RCC_BASE + AHB1ENR_OFFSET);

            // RCC AHB2 peripheral clock enable register (RCC_AHB2ENR)
            // Address offset: 0x34
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register AHB2ENR = new Register(STM32F4xx.RCC_BASE + AHB2ENR_OFFSET);

            // RCC AHB3 peripheral clock enable register (RCC_AHB3ENR)
            // Address offset: 0x38
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            Register AHB3ENR = new Register(STM32F4xx.RCC_BASE + AHB3ENR_OFFSET);

            // RCC APB1 peripheral clock enable register (RCC_APB1ENR)
            // Address offset: 0x40
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            // See: P.181
            Register APB1ENR = new Register(STM32F4xx.RCC_BASE + APB1ENR_OFFSET);

            // RCC APB2 peripheral clock enable register (RCC_APB2ENR)
            // Address offset: 0x44
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            // See: P.185
            Register APB2ENR = new Register(STM32F4xx.RCC_BASE + APB2ENR_OFFSET);

            // RCC AHB1 peripheral clock enable in low power mode register
            // (RCC_AHB1LPENR)
            // Address offset: 0x50
            // Reset value: 0x7E67 91FF
            // Access: no wait state, word, half-word and byte access.
            Register AHB1LPENR = new Register(STM32F4xx.RCC_BASE + AHB1LPENR_OFFSET);

            // RCC AHB2 peripheral clock enable in low power mode register
            // (RCC_AHB2LPENR)
            // Address offset: 0x54
            // Reset value: 0x0000 00F1
            // Access: no wait state, word, half-word and byte access.
            Register AHB2LPENR = new Register(STM32F4xx.RCC_BASE + AHB2LPENR_OFFSET);

            // RCC AHB3 peripheral clock enable in low power mode register
            // (RCC_AHB3LPENR)
            // Address offset: 0x58
            // Reset value: 0x0000 0001
            // Access: no wait state, word, half-word and byte access.
            Register AHB3LPENR = new Register(STM32F4xx.RCC_BASE + AHB3LPENR_OFFSET);

            // RCC APB1 peripheral clock enable in low power mode register
            // (RCC_APB1LPENR)
            // Address offset: 0x60
            // Reset value: 0x36FE C9FF
            // Access: no wait state, word, half-word and byte access.
            Register APB1LPENR = new Register(STM32F4xx.RCC_BASE + APB1LPENR_OFFSET);

            // RCC APB2 peripheral clock enabled in low power mode register
            // (RCC_APB2LPENR)
            // Address offset: 0x64
            // Reset value: 0x0007 5F33
            // Access: no wait state, word, half-word and byte access.
            Register APB2LPENR = new Register(STM32F4xx.RCC_BASE + APB2LPENR_OFFSET);

            // RCC Backup domain control register (RCC_BDCR)
            // Address offset: 0x70
            // Reset value: 0x0000 0000, reset by Backup domain reset.
            // Access: 0 ≤ wait state ≤ 3, word, half-word and byte access
            // Wait states are inserted in case of successive accesses to this register.
            Register BDCR = new Register(STM32F4xx.RCC_BASE + BDCR_OFFSET);

            // RCC clock control & status register (RCC_CSR)
            // Address offset: 0x74
            // Reset value: 0x0E00 0000, reset by system reset, except reset flags by power reset only.
            // Access: 0 ≤ wait state ≤ 3, word, half-word and byte access
            // Wait states are inserted in case of successive accesses to this register.
            Register CSR = new Register(STM32F4xx.RCC_BASE + CSR_OFFSET);

            // RCC spread spectrum clock generation register (RCC_SSCGR)
            // Address offset: 0x80
            // Reset value: 0x0000 0000
            // Access: no wait state, word, half-word and byte access.
            // The spread spectrum clock generation is available only for the main PLL.
            // The RCC_SSCGR register must be written either before the main PLL is enabled or after
            // the main PLL disabled.
            // Note: For full details about PLL spread spectrum clock generation (SSCG) characteristics, refer to
            // the “Electrical characteristics” section in your device datasheet.
            Register SSCGR = new Register(STM32F4xx.RCC_BASE + SSCGR_OFFSET);

            // RCC PLLI2S configuration register (RCC_PLLI2SCFGR)
            // Address offset: 0x84
            // Reset value: 0x2000 3000
            // Access: no wait state, word, half-word and byte access.
            // This register is used to configure the PLLI2S clock outputs according to the formulas:
            // f(VCO clock) = f(PLLI2S clock input) × (PLLI2SN / PLLM)
            // f(PLL I2S clock output) = f(VCO clock) / PLLI2SR
            Register PLLI2SCFGR = new Register(STM32F4xx.RCC_BASE + PLLI2SCFGR_OFFSET);

            #endregion

            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Set the Clock with RCC... See P.178 and P.238 and P.507 - Advanced-control timers (TIM1&TIM8)        //
            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Pulse lengths and waveform periods can be modulated from a few microseconds to several
            // milliseconds using the timer prescaler and the RCC clock controller prescalers.

            // Timer 8 CH1...
            if (PWMPin == CPU.Pin.Pin7_X3)
            {
                AHB1ENR.SetBits(1 << 2);  //   <--- GPIOCEN BIT 2 Enable...
                APB2ENR.SetBits(1 << 1);  //   <--- TIM8EN BIT 1 Enable...
            }
            // Timer 14 CH1...
            else if (PWMPin == CPU.Pin.Pin8_X3)
            {
                AHB1ENR.SetBits(1 << 0);  //   <--- GPIOAEN BIT 0 Enable...
                APB1ENR.SetBits(1 << 8);  //   <--- TIM14EN BIT 1 Enable...
            }
            // Timer 8 CH2...
            else if (PWMPin == CPU.Pin.Pin9_X3)
            {
                AHB1ENR.SetBits(1 << 2);  //   <--- GPIOCEN BIT 2 Enable...
                APB2ENR.SetBits(1 << 1);  //   <--- TIM8EN BIT 1 Enable...
            }
            // Timer 1 CH1...
            else if (PWMPin == CPU.Pin.Pin7_X4)
            {
                AHB1ENR.SetBits(1 << 0);  //   <--- GPIOAEN BIT 0 Enable...
                APB2ENR.SetBits(1 << 0);  //   <--- TIM1EN BIT 0 Enable...
            }
            // Timer 3 CH3...
            else if (PWMPin == CPU.Pin.Pin8_X4)
            {
                AHB1ENR.SetBits(1 << 1);  //   <--- GPIOBEN BIT 2 Enable...
                APB1ENR.SetBits(1 << 1);  //   <--- TIM3EN BIT 8 Enable...
            }
            // Timer 3 CH4...
            else if (PWMPin == CPU.Pin.Pin9_X4)
            {
                AHB1ENR.SetBits(1 << 1);  //   <--- GPIOBEN BIT 2 Enable...
                APB1ENR.SetBits(1 << 1);  //   <--- TIM3EN BIT 8 Enable...
            }

            #region Print Values...

            if (debug)
            {
                Debug.Print("*** RCC **********");
                // Print to Debug window current config...
                // Read Value after the change...
                Print("CFGR", CFGR); //   <--- Print Config...
                Print("AHB1ENR", AHB1ENR); //   <--- Print change...
                Print("APB2ENR", APB2ENR); //   <--- Print change...
            }

            #endregion
        }

        /// <summary>
        /// Configure GPIO Ports.
        /// Each of the 15 GPIOx Pins on each GPIO Port (GPIOA through I) is mapped to the MODER/AF/OSPEEDR/PUPDR and so on.
        /// This allows the control of each Pin, on each port to have Multiplexed Functions.
        /// Example: For GPIOC, MODER6 is mapped to Pin PA6, AF3 set on this Port in the AFRL (because the Pin is less than 7) sets TIM8 CH1.
        /// See: STM32F405VG User Manual P.278 and STM32F405VG P.62 for mappings.
        /// </summary>
        /// <param name="GPIOx_BASE">The Register BASE Address: 'STM32F4xx.GPIOC_BASE'</param>
        private void ConfGPIO()
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            // General-purpose I/O (GPIO) for STM32F405xx/07xx and STM32F415xx/17xx(RCC)                            //
            //////////////////////////////////////////////////////////////////////////////////////////////////////////

            #region OFFSET's...

            UInt32 MODER_OFFSET = 0x00;
            UInt32 OTYPER_OFFSET = 0x04;
            UInt32 OSPEEDR_OFFSET = 0x08;
            UInt32 PUPDR_OFFSET = 0x0C;
            UInt32 IDR_OFFSET = 0x10;
            UInt32 ODR_OFFSET = 0x14;
            UInt32 BSRR_OFFSET = 0x18;
            // UInt32 BSRRH_OFFSET = 0x1A; <--- ?
            UInt32 LCKR_OFFSET = 0x1C;

            #endregion

            #region Register's...

            // GPIO port mode register (GPIOx_MODER) (x = A..I/J/K)
            // Address offset: 0x00
            // Reset values:
            // • 0xA800 0000 for port A
            // • 0x0000 0280 for port B
            // • 0x0000 0000 for other ports
            Register MODER = new Register(this.GPIOx_BASE + MODER_OFFSET);

            // GPIO port output type register (GPIOx_OTYPER)
            // (x = A..I/J/K)
            // Address offset: 0x04
            // Reset value: 0x0000 0000
            Register OTYPER = new Register(this.GPIOx_BASE + OTYPER_OFFSET);

            // GPIO port output speed register (GPIOx_OSPEEDR)
            // (x = A..I/J/K)
            // Address offset: 0x08
            // Reset values:
            // • 0x0000 00C0 for port B
            // • 0x0000 0000 for other ports
            Register OSPEEDR = new Register(this.GPIOx_BASE + OSPEEDR_OFFSET);

            // GPIO port pull-up/pull-down register (GPIOx_PUPDR)
            // (x = A..I/J/K)
            // Address offset: 0x0C
            // Reset values:
            // • 0x6400 0000 for port A
            // • 0x0000 0100 for port B
            // • 0x0000 0000 for other ports
            Register PUPDR = new Register(this.GPIOx_BASE + PUPDR_OFFSET);

            // GPIO port input data register (GPIOx_IDR) (x = A..I/J/K)
            // Address offset: 0x10
            // Reset value: 0x0000 XXXX (where X means undefined)
            Register IDR = new Register(this.GPIOx_BASE + IDR_OFFSET);

            // GPIO port output data register (GPIOx_ODR) (x = A..I/J/K)
            // Address offset: 0x14
            // Reset value: 0x0000 0000
            Register ODR = new Register(this.GPIOx_BASE + ODR_OFFSET);

            // GPIO port bit set/reset register (GPIOx_BSRR) (x = A..I/J/K)
            // Address offset: 0x18
            // Reset value: 0x0000 0000
            Register BSRR = new Register(this.GPIOx_BASE + BSRR_OFFSET);

            // GPIO port configuration lock register (GPIOx_LCKR)
            // (x = A..I/J/K)
            // This register is used to lock the configuration of the port bits when a correct write sequence
            // is applied to bit 16 (LCKK). The value of bits [15:0] is used to lock the configuration of the
            // GPIO. During the write sequence, the value of LCKR[15:0] must not change. When the
            // LOCK sequence has been applied on a port bit, the value of this port bit can no longer be
            // modified until the next reset.
            // Note: A specific write sequence is used to write to the GPIOx_LCKR register. Only word access
            // (32-bit long) is allowed during this write sequence.
            // Each lock bit freezes a specific configuration register (control and alternate function
            // registers).
            // Address offset: 0x1C
            // Reset value: 0x0000 0000
            // Access: 32-bit word only, read/write register
            Register LCKR = new Register(this.GPIOx_BASE + LCKR_OFFSET);

            // Alternate Functions...
            // GPIO alternate function low register (GPIOx_AFRL) (x = A..I/J/K) - AFR[0]
            // Address offset: 0x20
            // Reset value: 0x0000 0000
            // GPIO alternate function high register (GPIOx_AFRH) - AFR[1]
            // (x = A..I/J/K)
            // Address offset: 0x24
            // Reset value: 0x0000 0000
            Register[] AFR = new Register[] { new Register(this.GPIOx_BASE + 0x20), new Register(this.GPIOx_BASE + 0x24) };

            #endregion

            // Timer 8 CH1...
            if (PWMPin == CPU.Pin.Pin7_X3)
            {
                #region CPU.Pin.Pin7_X3

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.283 - GPIO alternate function low register
                // AFRLy: Alternate function selection for port x bit y (y = 0..7)
                // These bits are written by software to configure alternate function I/Os
                // AFRLy selection:
                // 0000: AF0 (system)
                // 0001: AF1 (TIM1/TIM2)
                // 0010: AF2 (TIM3..5)
                // 0011: AF3 (TIM8..11)
                // 0100: AF4 (I2C1..3)
                // 0101: AF5 (SPI1/SPI2)
                // 0110: AF6 (SPI3)
                // 0111: AF7 (USART1..3)
                // 1000: AF8 (USART4..6)
                // 1001: AF9 (CAN1/CAN2, TIM12..14)
                // 1010: AF10 (OTG_FS, OTG_HS)
                // 1011: AF11 (ETH)
                // 1100: AF12 (FSMC, SDIO, OTG_HS(1))
                // 1101: AF13 (DCMI)
                // 1110: AF14
                // 1111: AF15 (EVENTOUT)
                // See: Figure 26. Selecting an alternate function
                // Also See: Table 9. Alternate function mapping of STM32F405VG.pdf
                if (AFR[0].Read() == 0)
                {
                    AFR[0].Write(0x3000000);
                }
                else
                {
                    AFR[0].SetBits(3 << 24); //   <--- Set Bits 24 and 25 to 11 - AFRL Register - Enables AF3 on AFRL6 this Maps to Pin GPIOC 6 (PC6) which is Pin 5 on X3...
                }

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Configure CPU Pins... See: P.271 - I/O port control registers...                                     //
                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Each of the GPIOs has four 32-bit memory-mapped control registers (GPIOx_MODER,
                // GPIOx_OTYPER, GPIOx_OSPEEDR, GPIOx_PUPDR) to configure up to 16 I/Os. The
                // GPIOx_MODER register is used to select the I/O direction (input, output, AF, analog).
                // MODERy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O direction mode.
                // 00: Input (reset state)
                // 01: General purpose output mode
                // 10: Alternate function mode
                // 11: Analog mode
                // MODER.ClearBits(1 << 12);
                MODER.SetBits(2 << 12); //   <--- Set Bits 12 and 13 to 10 - MODER6 this Maps to Pin GPIOC 6 (PC6) which is Pin 5 on X3...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output type register
                // Bits 15:0 OTy: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the output type of the I/O port.
                // 0: Output push-pull (reset state)
                // 1: Output open-drain
                OTYPER.ClearBits(1 << 6); //   <--- Set Bit 6 to 0 - OT6 this Maps to Pin GPIOC 6 (PC6) which is Pin 5 on X3...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output speed register
                // OSPEEDRy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O output speed.
                // 00: Low speed
                // 01: Medium speed
                // 10: Fast speed
                // 11: High speed
                // Note: Refer to the
                OSPEEDR.SetBits(3 << 12); //   <--- Set Bits 12 and 13 to 11 - OSPEEDR6 this Maps to Pin GPIOC 6 (PC6) which is Pin 5 on X3...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.280 - GPIO port pull-up/pull-down register (GPIOx_PUPDR) 
                // (x = A..I/J/K) - Address offset: 0x0C
                // Reset values: 
                // • 0x6400 0000 for port A 
                // • 0x0000 0100 for port B 
                // • 0x0000 0000 for other ports
                // The GPIOx_PUPDR register is used to select the pull-up/pull-down whatever the I/O direction.
                PUPDR.ClearBits(3 << 12); //   <--- Set Bits 12 and 13 to 11 - PUPDR6 this Maps to Pin GPIOC 6 (PC6) which is Pin 5 on X3...

                #endregion
            }
            // Timer 14 CH1...
            else if (PWMPin == CPU.Pin.Pin8_X3)
            {
                #region CPU.Pin.Pin8_X3

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.283 - GPIO alternate function low register
                // AFRLy: Alternate function selection for port x bit y (y = 0..7)
                // These bits are written by software to configure alternate function I/Os
                // AFRLy selection:
                // 0000: AF0 (system)
                // 0001: AF1 (TIM1/TIM2)
                // 0010: AF2 (TIM3..5)
                // 0011: AF3 (TIM8..11)
                // 0100: AF4 (I2C1..3)
                // 0101: AF5 (SPI1/SPI2)
                // 0110: AF6 (SPI3)
                // 0111: AF7 (USART1..3)
                // 1000: AF8 (USART4..6)
                // 1001: AF9 (CAN1/CAN2, TIM12..14)
                // 1010: AF10 (OTG_FS, OTG_HS)
                // 1011: AF11 (ETH)
                // 1100: AF12 (FSMC, SDIO, OTG_HS(1))
                // 1101: AF13 (DCMI)
                // 1110: AF14
                // 1111: AF15 (EVENTOUT)
                // See: Figure 26. Selecting an alternate function
                // Also See: Table 9. Alternate function mapping of STM32F405VG.pdf
                // AFR[0].ClearBits(0xF0000000);
                AFR[0].SetBits(0x98000000);//   <--- Set Bits 32 to 28 - AFRL Register - Enables AF1 on AFRL7 this Maps to Pin GPIOA 7 (PA7) which is Pin 8 on X3...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Configure CPU Pins... See: P.271 - I/O port control registers...                                     //
                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Each of the GPIOs has four 32-bit memory-mapped control registers (GPIOx_MODER,
                // GPIOx_OTYPER, GPIOx_OSPEEDR, GPIOx_PUPDR) to configure up to 16 I/Os. The
                // GPIOx_MODER register is used to select the I/O direction (input, output, AF, analog).
                // MODERy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O direction mode.
                // 00: Input (reset state)
                // 01: General purpose output mode
                // 10: Alternate function mode
                // 11: Analog mode
                // MODER.ClearBits(1 << 12);
                MODER.SetBits(2 << 14); //   <--- Set Bits 14 and 15 to 10 - MODER7 this Maps to Pin GPIOA 7 (PA7) which is Pin 8 on X3...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output type register
                // Bits 15:0 OTy: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the output type of the I/O port.
                // 0: Output push-pull (reset state)
                // 1: Output open-drain
                OTYPER.ClearBits(1 << 7); //   <--- Set Bit 7 to 0 - OT7 this Maps to Pin GPIOA 7 (PA7) which is Pin 8 on X3...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output speed register
                // OSPEEDRy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O output speed.
                // 00: Low speed
                // 01: Medium speed
                // 10: Fast speed
                // 11: High speed
                // Note: Refer to the
                OSPEEDR.SetBits(3 << 14); //   <--- Set Bits 6 and 7 to 11 - OSPEEDR3 this Maps to Pin GPIOA 7 (PA7) which is Pin 8 on X3...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.280 - GPIO port pull-up/pull-down register (GPIOx_PUPDR) 
                // (x = A..I/J/K) - Address offset: 0x0C
                // Reset values: 
                // • 0x6400 0000 for port A 
                // • 0x0000 0100 for port B 
                // • 0x0000 0000 for other ports
                // The GPIOx_PUPDR register is used to select the pull-up/pull-down whatever the I/O direction.
                PUPDR.ClearBits(3 << 14); //   <--- Set Bits 14 and 15 to 11 - PUPDR3 this Maps to Pin GPIOA 7 (PA7) which is Pin 8 on X3...

                #endregion
            }
            // Timer 8 CH2...
            else if (PWMPin == CPU.Pin.Pin9_X3)
            {
                #region CPU.Pin.Pin9_X3

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.283 - GPIO alternate function low register
                // AFRLy: Alternate function selection for port x bit y (y = 0..7)
                // These bits are written by software to configure alternate function I/Os
                // AFRLy selection:
                // 0000: AF0 (system)
                // 0001: AF1 (TIM1/TIM2)
                // 0010: AF2 (TIM3..5)
                // 0011: AF3 (TIM8..11)
                // 0100: AF4 (I2C1..3)
                // 0101: AF5 (SPI1/SPI2)
                // 0110: AF6 (SPI3)
                // 0111: AF7 (USART1..3)
                // 1000: AF8 (USART4..6)
                // 1001: AF9 (CAN1/CAN2, TIM12..14)
                // 1010: AF10 (OTG_FS, OTG_HS)
                // 1011: AF11 (ETH)
                // 1100: AF12 (FSMC, SDIO, OTG_HS(1))
                // 1101: AF13 (DCMI)
                // 1110: AF14
                // 1111: AF15 (EVENTOUT)
                // See: Figure 26. Selecting an alternate function
                // Also See: Table 9. Alternate function mapping of STM32F405VG.pdf
                if (AFR[0].Read() == 0)
                {
                    AFR[0].Write(805306368); // 805306368 // 0x30000000
                }
                else
                {
                    AFR[0].SetBits(3 << 28); //   <--- Set Bits 24 and 25 to 11 - AFRL Register - Enables AF3 on AFRL6 this Maps to Pin GPIOC 6 (PC6) which is Pin 9 on X3...
                }

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Configure CPU Pins... See: P.271 - I/O port control registers...                                     //
                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Each of the GPIOs has four 32-bit memory-mapped control registers (GPIOx_MODER,
                // GPIOx_OTYPER, GPIOx_OSPEEDR, GPIOx_PUPDR) to configure up to 16 I/Os. The
                // GPIOx_MODER register is used to select the I/O direction (input, output, AF, analog).
                // MODERy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O direction mode.
                // 00: Input (reset state)
                // 01: General purpose output mode
                // 10: Alternate function mode
                // 11: Analog mode
                MODER.SetBits(2 << 14); //   <--- Set Bits 14 and 15 to 10 - MODER7 this Maps to Pin GPIOC 7 (PC7) which is Pin 9 on X3...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output type register
                // Bits 15:0 OTy: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the output type of the I/O port.
                // 0: Output push-pull (reset state)
                // 1: Output open-drain
                OTYPER.ClearBits(1 << 7); //   <--- Set Bit 7 to 0 - OT6 this Maps to Pin GPIOC 6 (PC6) which is Pin 5 on X3...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output speed register
                // OSPEEDRy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O output speed.
                // 00: Low speed
                // 01: Medium speed
                // 10: Fast speed
                // 11: High speed
                // Note: Refer to the
                OSPEEDR.SetBits(3 << 9); //   <--- Set Bits 14 and 15 to 11 - OSPEEDR7 this Maps to Pin GPIOC 7 (PC7) which is Pin 9 on X3...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.280 - GPIO port pull-up/pull-down register (GPIOx_PUPDR) 
                // (x = A..I/J/K) - Address offset: 0x0C
                // Reset values: 
                // • 0x6400 0000 for port A 
                // • 0x0000 0100 for port B 
                // • 0x0000 0000 for other ports
                // The GPIOx_PUPDR register is used to select the pull-up/pull-down whatever the I/O direction.
                PUPDR.ClearBits(3 << 14); //   <--- Set Bits 14 and 15 to 11 - PUPDR7 this Maps to Pin GPIOC 7 (PC7) which is Pin 9 on X3...

                #endregion
            }
            // Timer 1 CH1...
            else if (PWMPin == CPU.Pin.Pin7_X4)
            {
                #region CPU.Pin.Pin7_X4

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.283 - GPIO alternate function low register
                // AFRLy: Alternate function selection for port x bit y (y = 0..7)
                // These bits are written by software to configure alternate function I/Os
                // AFRLy selection:
                // 0000: AF0 (system)
                // 0001: AF1 (TIM1/TIM2)
                // 0010: AF2 (TIM3..5)
                // 0011: AF3 (TIM8..11)
                // 0100: AF4 (I2C1..3)
                // 0101: AF5 (SPI1/SPI2)
                // 0110: AF6 (SPI3)
                // 0111: AF7 (USART1..3)
                // 1000: AF8 (USART4..6)
                // 1001: AF9 (CAN1/CAN2, TIM12..14)
                // 1010: AF10 (OTG_FS, OTG_HS)
                // 1011: AF11 (ETH)
                // 1100: AF12 (FSMC, SDIO, OTG_HS(1))
                // 1101: AF13 (DCMI)
                // 1110: AF14
                // 1111: AF15 (EVENTOUT)
                // See: Figure 26. Selecting an alternate function
                // Also See: Table 9. Alternate function mapping of STM32F405VG.pdf
                AFR[1].SetBits(1 << 0); //   <--- Set Bits 0 to 1 - AFRH Register - Enables AF1 on AFRL8 this Maps to Pin GPIOA 8 (PA8) which is Pin 7 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Configure CPU Pins... See: P.271 - I/O port control registers...                                     //
                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Each of the GPIOs has four 32-bit memory-mapped control registers (GPIOx_MODER,
                // GPIOx_OTYPER, GPIOx_OSPEEDR, GPIOx_PUPDR) to configure up to 16 I/Os. The
                // GPIOx_MODER register is used to select the I/O direction (input, output, AF, analog).
                // MODERy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O direction mode.
                // 00: Input (reset state)
                // 01: General purpose output mode
                // 10: Alternate function mode
                // 11: Analog mode
                // MODER.ClearBits(1 << 12);
                MODER.SetBits(2 << 16); //   <--- Set Bits 16 and 17 to 10 - MODER8 this Maps to Pin GPIOA 8 (PA8) which is Pin 7 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output type register
                // Bits 15:0 OTy: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the output type of the I/O port.
                // 0: Output push-pull (reset state)
                // 1: Output open-drain
                OTYPER.ClearBits(1 << 7); //   <--- Set Bit 7 to 0 - OT8 this Maps to Pin GPIOA 8 (PA8) which is Pin 7 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output speed register
                // OSPEEDRy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O output speed.
                // 00: Low speed
                // 01: Medium speed
                // 10: Fast speed
                // 11: High speed
                // Note: Refer to the
                OSPEEDR.SetBits(3 << 16); //   <--- Set Bits 16 and 17 to 11 - OSPEEDR8 this Maps to Pin GPIOA 8 (PA8) which is Pin 7 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.280 - GPIO port pull-up/pull-down register (GPIOx_PUPDR) 
                // (x = A..I/J/K) - Address offset: 0x0C
                // Reset values: 
                // • 0x6400 0000 for port A 
                // • 0x0000 0100 for port B 
                // • 0x0000 0000 for other ports
                // The GPIOx_PUPDR register is used to select the pull-up/pull-down whatever the I/O direction.
                PUPDR.ClearBits(3 << 16); //   <--- Set Bits 16 and 17 to 11 - PUPDR8 this Maps to Pin GPIOA 8 (PA8) which is Pin 7 on X4...

                #endregion
            }
            // Timer 3 CH3...
            else if (PWMPin == CPU.Pin.Pin8_X4)
            {
                #region CPU.Pin.Pin8_X4

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.283 - GPIO alternate function low register
                // AFRLy: Alternate function selection for port x bit y (y = 0..7)
                // These bits are written by software to configure alternate function I/Os
                // AFRLy selection:
                // 0000: AF0 (system)
                // 0001: AF1 (TIM1/TIM2)
                // 0010: AF2 (TIM3..5)
                // 0011: AF3 (TIM8..11)
                // 0100: AF4 (I2C1..3)
                // 0101: AF5 (SPI1/SPI2)
                // 0110: AF6 (SPI3)
                // 0111: AF7 (USART1..3)
                // 1000: AF8 (USART4..6)
                // 1001: AF9 (CAN1/CAN2, TIM12..14)
                // 1010: AF10 (OTG_FS, OTG_HS)
                // 1011: AF11 (ETH)
                // 1100: AF12 (FSMC, SDIO, OTG_HS(1))
                // 1101: AF13 (DCMI)
                // 1110: AF14
                // 1111: AF15 (EVENTOUT)
                // See: Figure 26. Selecting an alternate function
                // Also See: Table 9. Alternate function mapping of STM32F405VG.pdf
                AFR[0].ClearBits(0xF << 0); //   <--- Set Bits 4 and 5 to 10 - AFRL Register - Enables AF2 on AFRL1 this Maps to Pin GPIOB 0 (PB0) which is Pin 8 on X4...
                AFR[0].SetBits(2 << 0);

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Configure CPU Pins... See: P.271 - I/O port control registers...                                     //
                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Each of the GPIOs has four 32-bit memory-mapped control registers (GPIOx_MODER,
                // GPIOx_OTYPER, GPIOx_OSPEEDR, GPIOx_PUPDR) to configure up to 16 I/Os. The
                // GPIOx_MODER register is used to select the I/O direction (input, output, AF, analog).
                // MODERy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O direction mode.
                // 00: Input (reset state)
                // 01: General purpose output mode
                // 10: Alternate function mode
                // 11: Analog mode
                MODER.SetBits(2 << 0); //   <--- Set Bits 0 and 1 to 10 - MODER0 this Maps to Pin GPIOB 0 (PB0) which is Pin 8 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output type register
                // Bits 15:0 OTy: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the output type of the I/O port.
                // 0: Output push-pull (reset state)
                // 1: Output open-drain
                OTYPER.ClearBits(1 << 0); //   <--- Set Bit 0 to 0 - OT0 this Maps to Pin GPIOB 0 (PB0) which is Pin 8 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output speed register
                // OSPEEDRy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O output speed.
                // 00: Low speed
                // 01: Medium speed
                // 10: Fast speed
                // 11: High speed
                // Note: Refer to the
                OSPEEDR.SetBits(3 << 0); //   <--- Set Bits 0 and 1 to 11 - OSPEEDR2 this Maps to Pin GPIOB 0 (PB0) which is Pin 8 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.280 - GPIO port pull-up/pull-down register (GPIOx_PUPDR) 
                // (x = A..I/J/K) - Address offset: 0x0C
                // Reset values: 
                // • 0x6400 0000 for port A 
                // • 0x0000 0100 for port B 
                // • 0x0000 0000 for other ports
                // The GPIOx_PUPDR register is used to select the pull-up/pull-down whatever the I/O direction.
                PUPDR.ClearBits(3 << 0); //   <--- Set Bits 0 and 1 to 11 - PUPDR1 this Maps to Pin GPIOB 0 (PB0) which is Pin 8 on X4...

                #endregion
            }
            // Timer 3 CH4...
            else if (PWMPin == CPU.Pin.Pin9_X4)
            {
                #region CPU.Pin.Pin9_X4

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.283 - GPIO alternate function low register
                // AFRLy: Alternate function selection for port x bit y (y = 0..7)
                // These bits are written by software to configure alternate function I/Os
                // AFRLy selection:
                // 0000: AF0 (system)
                // 0001: AF1 (TIM1/TIM2)
                // 0010: AF2 (TIM3..5)
                // 0011: AF3 (TIM8..11)
                // 0100: AF4 (I2C1..3)
                // 0101: AF5 (SPI1/SPI2)
                // 0110: AF6 (SPI3)
                // 0111: AF7 (USART1..3)
                // 1000: AF8 (USART4..6)
                // 1001: AF9 (CAN1/CAN2, TIM12..14)
                // 1010: AF10 (OTG_FS, OTG_HS)
                // 1011: AF11 (ETH)
                // 1100: AF12 (FSMC, SDIO, OTG_HS(1))
                // 1101: AF13 (DCMI)
                // 1110: AF14
                // 1111: AF15 (EVENTOUT)
                // See: Figure 26. Selecting an alternate function
                // Also See: Table 9. Alternate function mapping of STM32F405VG.pdf
                AFR[0].ClearBits(0xF << 4);
                AFR[0].SetBits(2 << 4); //   <--- Set Bits 4 and 5 to 10 - AFRL Register - Enables AF2 on AFRL1 this Maps to Pin GPIOB 1 (PB1) which is Pin 9 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Configure CPU Pins... See: P.271 - I/O port control registers...                                     //
                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // Each of the GPIOs has four 32-bit memory-mapped control registers (GPIOx_MODER,
                // GPIOx_OTYPER, GPIOx_OSPEEDR, GPIOx_PUPDR) to configure up to 16 I/Os. The
                // GPIOx_MODER register is used to select the I/O direction (input, output, AF, analog).
                // MODERy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O direction mode.
                // 00: Input (reset state)
                // 01: General purpose output mode
                // 10: Alternate function mode
                // 11: Analog mode
                // MODER.ClearBits(1 << 12);
                MODER.SetBits(2 << 2); //   <--- Set Bits 2 and 3 to 10 - MODER1 this Maps to Pin GPIOB 1 (PB1) which is Pin 9 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output type register
                // Bits 15:0 OTy: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the output type of the I/O port.
                // 0: Output push-pull (reset state)
                // 1: Output open-drain
                OTYPER.ClearBits(1 << 1); //   <--- Set Bit 1 to 0 - OT1 this Maps to Pin GPIOB 1 (PB1) which is Pin 9 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.279 - GPIO port output speed register
                // OSPEEDRy[1:0]: Port x configuration bits (y = 0..15)
                // These bits are written by software to configure the I/O output speed.
                // 00: Low speed
                // 01: Medium speed
                // 10: Fast speed
                // 11: High speed
                // Note: Refer to the
                OSPEEDR.SetBits(3 << 2); //   <--- Set Bits 2 and 3 to 11 - OSPEEDR1 this Maps to Pin GPIOB 1 (PB1) which is Pin 9 on X4...

                //////////////////////////////////////////////////////////////////////////////////////////////////////////
                // See: P.280 - GPIO port pull-up/pull-down register (GPIOx_PUPDR) 
                // (x = A..I/J/K) - Address offset: 0x0C
                // Reset values: 
                // • 0x6400 0000 for port A 
                // • 0x0000 0100 for port B 
                // • 0x0000 0000 for other ports
                // The GPIOx_PUPDR register is used to select the pull-up/pull-down whatever the I/O direction.
                PUPDR.ClearBits(3 << 1); //   <--- Set Bits 14 and 15 to 11 - PUPDR1 this Maps to Pin GPIOB 3 (PB1) which is Pin 5 on X3...

                #endregion
            }

            #region Print Values...

            if (debug)
            {
                // Print to Debug window current config...
                Debug.Print("*** GPIO *********");

                Print("AFR[0]", AFR[0]); //   <--- Print change...
                Print("AFR[1]", AFR[1]); //   <--- Print change...
                Print("MODER", MODER); //   <--- Print change...
                Print("OTYPER", OTYPER); //   <--- Print change...
                Print("OSPEEDR", OSPEEDR); //   <--- Print change...
                Print("PUPDR", PUPDR); //   <--- Print change...
            }

            #endregion
        }

        /// <summary>
        /// Timer Configuration.
        /// </summary>
        /// <param name="TIMx_BASE">The Register BASE Address: 'STM32F4xx.TIM8_BASE'</param>
        private void ConfTimer()
        {
            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            // General-purpose I/O (GPIO) for STM32F405xx/07xx and STM32F415xx/17xx(RCC)                            //
            //////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Good Resources:
            // Table 42. DMA1 request mapping - See: P.304
            // Table 43. DMA2 request mapping - See: P.305
            // Advanced-control timers (TIM1&TIM8) - See: P.507

            #region OFFSET's...

            UInt32 CR1_OFFSET = 0x00;
            UInt32 CR2_OFFSET = 0x04;
            UInt32 SMCR_OFFSET = 0x08;
            UInt32 DIER_OFFSET = 0x0C;
            UInt32 SR_OFFSET = 0x10;
            UInt32 EGR_OFFSET = 0x14;
            UInt32 CCMR1_OFFSET = 0x18;
            UInt32 CCMR2_OFFSET = 0x1C;
            UInt32 CCER_OFFSET = 0x20;
            UInt32 CNT_OFFSET = 0x24;
            UInt32 PSC_OFFSET = 0x28;
            UInt32 ARR_OFFSET = 0x2C;
            UInt32 RCR_OFFSET = 0x30;
            UInt32 CCR1_OFFSET = 0x34;
            UInt32 CCR2_OFFSET = 0x38;
            UInt32 CCR3_OFFSET = 0x3C;
            UInt32 CCR4_OFFSET = 0x40;
            UInt32 BDTR_OFFSET = 0x44;
            UInt32 DCR_OFFSET = 0x48;
            UInt32 DMAR_OFFSET = 0x4C;
            UInt32 OR_OFFSET = 0x50;

            #endregion

            #region Registers...

            // TIM1&TIM8 control register 1 (TIMx_CR1)
            // Address offset: 0x00
            // Reset value: 0x0000
            Register CR1 = new Register(this.TIMx_BASE + CR1_OFFSET);

            // TIM1&TIM8 control register 2 (TIMx_CR2)
            // Address offset: 0x04
            // Reset value: 0x0000
            Register CR2 = new Register(this.TIMx_BASE + CR2_OFFSET);

            // TIM1&TIM8 slave mode control register (TIMx_SMCR)
            // Address offset: 0x08
            // Reset value: 0x0000
            Register SMCR = new Register(this.TIMx_BASE + SMCR_OFFSET);

            // TIM1&TIM8 DMA/interrupt enable register (TIMx_DIER)
            // Address offset: 0x0C
            // Reset value: 0x0000
            Register DIER = new Register(this.TIMx_BASE + DIER_OFFSET);

            // TIM1&TIM8 status register (TIMx_SR)
            // Address offset: 0x10
            // Reset value: 0x0000
            Register SR = new Register(this.TIMx_BASE + SR_OFFSET);

            // TIM1&TIM8 event generation register (TIMx_EGR)
            // Address offset: 0x14
            // Reset value: 0x0000
            Register EGR = new Register(this.TIMx_BASE + EGR_OFFSET);

            // TIM1&TIM8 capture/compare mode register 1 (TIMx_CCMR1)
            // Address offset: 0x18
            // Reset value: 0x0000
            Register CCMR1 = new Register(this.TIMx_BASE + CCMR1_OFFSET);

            // TIM1&TIM8 capture/compare mode register 2 (TIMx_CCMR2)
            // Address offset: 0x1C
            // Reset value: 0x0000
            // Refer to the above CCMR1 register description.
            Register CCMR2 = new Register(this.TIMx_BASE + CCMR2_OFFSET);

            // TIM1&TIM8 capture/compare enable register (TIMx_CCER)
            // Address offset: 0x20
            // Reset value: 0x0000
            Register CCER = new Register(this.TIMx_BASE + CCER_OFFSET);

            // TIM1&TIM8 counter (TIMx_CNT)
            // Address offset: 0x24
            // Reset value: 0x0000
            Register CNT = new Register(this.TIMx_BASE + CNT_OFFSET);

            // TIM1&TIM8 prescaler (TIMx_PSC)
            // Address offset: 0x28
            // Reset value: 0x0000
            Register PSC = new Register(this.TIMx_BASE + PSC_OFFSET);

            // TIM1&TIM8 auto-reload register (TIMx_ARR)
            // Address offset: 0x2C
            // Reset value: 0x0000
            Register ARR = new Register(this.TIMx_BASE + ARR_OFFSET);

            // TIM1&TIM8 repetition counter register (TIMx_RCR)
            // Address offset: 0x30
            // Reset value: 0x0000
            Register RCR = new Register(this.TIMx_BASE + RCR_OFFSET);

            // TIM1&TIM8 capture/compare register 1 (TIMx_CCR1)
            // Address offset: 0x34
            // Reset value: 0x0000
            Register CCR1 = new Register(this.TIMx_BASE + CCR1_OFFSET);

            // TIM1&TIM8 capture/compare register 2 (TIMx_CCR2)
            // Address offset: 0x38
            // Reset value: 0x0000
            Register CCR2 = new Register(this.TIMx_BASE + CCR2_OFFSET);

            // TIM1&TIM8 capture/compare register 3 (TIMx_CCR3)
            // Address offset: 0x3C
            // Reset value: 0x0000
            Register CCR3 = new Register(this.TIMx_BASE + CCR3_OFFSET);

            // TIM1&TIM8 capture/compare register 4 (TIMx_CCR4)
            // Address offset: 0x40
            // Reset value: 0x0000
            Register CCR4 = new Register(this.TIMx_BASE + CCR4_OFFSET);

            // TIM1&TIM8 break and dead-time register (TIMx_BDTR)
            // Address offset: 0x44
            // Reset value: 0x0000
            Register BDTR = new Register(this.TIMx_BASE + BDTR_OFFSET);

            // TIM1&TIM8 DMA control register (TIMx_DCR)
            // Address offset: 0x48
            // Reset value: 0x0000
            Register DCR = new Register(this.TIMx_BASE + DCR_OFFSET);

            // TIM1&TIM8 DMA address for full transfer (TIMx_DMAR)
            // Address offset: 0x4C
            // Reset value: 0x0000
            Register DMAR = new Register(this.TIMx_BASE + DMAR_OFFSET);

            #endregion

            ARR.Write(ARRValue);

            PSC.Write(PSCValue);

            #region Set the Value for the Channel on the Pin...

            /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Channels refer to the CCRx Registers. Some Timers can have 4 Channels per Timer. CCR1, CCR2, CCR3, and CCR4...          //
            // See: P.8 - AN4013 - http://www.st.com/web/en/resource/technical/document/application_note/DM00042534.pdf                //
            /////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            if (PWMChannel == Channel.CH1)
            {
                CCR1.Write(CCRxValue);
            }
            else if (PWMChannel == Channel.CH2)
            {
                CCR2.Write(CCRxValue);
            }
            else if (PWMChannel == Channel.CH3)
            {
                CCR3.Write(CCRxValue);
            }
            else if (PWMChannel == Channel.CH4)
            {
                CCR4.Write(CCRxValue);
            }

            #endregion

            #region Pin Configuration...

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bit 0 CEN: Counter enable                                                                                              //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // 0: Counter disabled
            // 1: Counter enabled
            // Note: External clock, gated mode and encoder mode can work only if the CEN bit has been
            // previously set by software. However trigger mode can set the CEN bit automatically by hardware.
            // CEN is cleared automatically in one-pulse mode, when an update event occurs.
            CR1.ClearBits(1 << 0); //   <--- Bit 0 CEN: Counter disable

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // See: P.548 - Bits 6:5 CMS[1:0]: Center-aligned mode selection                                                          //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            CenterAlignModeSelection(this.TIMx_BASE, CPU.TriggerMode.EdgeAlignedMode);

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Set Output Compare Mode - Channel Mode Selection (CH3 == OC3)                                                          //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            SetOutputCompareMode(this.TIMx_BASE, this.PWMChannel, CPU.OCxMode.PWMMode1);

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Invert the CPU Pin - This Toggles the current bits in the CCxP Registers                                               //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            InvertCPUPin(this.TIMx_BASE, this.PWMChannel);

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bits 1:0 CC1S: Capture/Compare 1 selection                                                                             //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            CompareEnable(this.TIMx_BASE, this.PWMChannel);

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bit 0 CEN: Counter enable                                                                                              //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // 0: Counter disabled
            // 1: Counter enabled
            // Note: External clock, gated mode and encoder mode can work only if the CEN bit has been
            // previously set by software. However trigger mode can set the CEN bit automatically by hardware.
            // CEN is cleared automatically in one-pulse mode, when an update event occurs.
            CR1.SetBits(1 << 0); //   <--- Bit 0 CEN: Counter enable

            #endregion

            #region Print Values...

            if (debug)
            {
                Debug.Print("*** Timer ********");
                // Print to Debug window new config...
                Print("CR1", CR1);
                Print("CR2", CR2);
                Print("SMCR", SMCR);
                Print("DIER", DIER);
                Print("SR", SR);
                Print("EGR", EGR);
                Print("CCMR1", CCMR1);
                Print("CCMR2", CCMR2);
                Print("CCER", CCER);
                Print("CNT", CNT);
                Print("PSC", PSC);
                Print("ARR", ARR);
                Print("RCR", RCR);
                Print("CCR1", CCR1);
                Print("CCR2", CCR2);
                Print("CCR3", CCR3);
                Print("CCR4", CCR4);
                Print("BDTR", BDTR);
                Print("DCR", DCR);
                Print("DMAR", DMAR);
                Debug.Print("******************");
            }

            #endregion
        }

        #endregion

        private uint CycleOffSet(UInt32 TIMx_BASE, int DegreeOfOffSet)
        {
            UInt32 ARR_OFFSET = 0x2C;
            Register ARR = new Register(TIMx_BASE + ARR_OFFSET);

            uint ARRegister = ARR.Read();

            // Period is the seconds/cycle
            // If a coil makes 8 vibrational cycles in 4 seconds, then the frequency is 2 Hz (8 cycles/4 s = 2 cycles/s).

            uint OffSet;

            if (DegreeOfOffSet > 360)
            {
                DegreeOfOffSet = 360;
            }

            if (DegreeOfOffSet == 0)
            {
                return 0;
            }

            uint TPerDegree = (this.Period / 360);

            OffSet = (uint)(TPerDegree * DegreeOfOffSet);

            if (debug)
            {
                Debug.Print("*** OffSet *******");
                // Print to Debug window new config...
                Debug.Print("OffSet: " + OffSet);
                Debug.Print("Period: " + this.Period);
            }


            return OffSet;
        }

        public void Enable(CPU.Pin pin)
        {
            try
            {
                // Configure current Pin...
                SetPinPort(pin);
            }
            catch
            {
                Debug.Print("Error setting the Pin Info in Enable...");
            }

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // See: P.602 - 18.3.14 Timers and external trigger synchronization                                                       //
            // See: P.606 - Using one timer to enable another timer                                                                   //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

            UInt32 CR1_OFFSET = 0x00;
            Register CR1 = new Register(this.TIMx_BASE + CR1_OFFSET);

            UInt32 CR2_OFFSET = 0x04;
            Register CR2 = new Register(this.TIMx_BASE + CR2_OFFSET);

            UInt32 CNT_OFFSET = 0x24;
            Register CNT = new Register(this.TIMx_BASE + CNT_OFFSET);

            UInt32 EGR_OFFSET = 0x14;
            Register EGR = new Register(this.TIMx_BASE + EGR_OFFSET);

            UInt32 SMCR_OFFSET = 0x08;
            Register SMCR = new Register(this.TIMx_BASE + SMCR_OFFSET);

            CR1.SetBits(1 << 0);
        }

        public void Disable(CPU.Pin pin)
        {
            try
            {
                // Configure current Pin...
                SetPinPort(pin);
            }
            catch
            {
                Debug.Print("Error setting the Pin Info in Disable...");
            }

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // See: P.602 - 18.3.14 Timers and external trigger synchronization                                                       //
            // See: P.606 - Using one timer to enable another timer                                                                   //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

            UInt32 CR1_OFFSET = 0x00;
            Register CR1 = new Register(this.TIMx_BASE + CR1_OFFSET);

            UInt32 CR2_OFFSET = 0x04;
            Register CR2 = new Register(this.TIMx_BASE + CR2_OFFSET);

            UInt32 CNT_OFFSET = 0x24;
            Register CNT = new Register(this.TIMx_BASE + CNT_OFFSET);

            UInt32 EGR_OFFSET = 0x14;
            Register EGR = new Register(this.TIMx_BASE + EGR_OFFSET);

            UInt32 SMCR_OFFSET = 0x08;
            Register SMCR = new Register(this.TIMx_BASE + SMCR_OFFSET);

            CR1.ClearBits(1 << 0);
        }

        public void RebootBoard()
        {
            Microsoft.SPOT.Hardware.PowerState.RebootDevice(false);
        }


        #region Register Configuration Methods...

        private void CenterAlignModeSelection(UInt32 TIMx_BASE, CPU.TriggerMode mode)
        {
            UInt32 CR1_OFFSET = 0x00;
            Register CR1 = new Register(TIMx_BASE + CR1_OFFSET);

            CR1.ClearBits(3 << 5);
            CR1.SetBits((uint)mode << 5); //   <--- Bits 6:5 CMS: Center-aligned mode selection
        }

        // Capture/Compare Mode Register Configuration...
        // See: P.559 and P.562 for a description: RM0090
        public void SetOutputCompareMode(UInt32 TIMx_BASE, Channel channel, CPU.OCxMode mode)
        {
            UInt32 CCMR1_OFFSET = 0x18;
            UInt32 CCMR2_OFFSET = 0x1C;
            Register CCMR1 = new Register(TIMx_BASE + CCMR1_OFFSET);
            Register CCMR2 = new Register(TIMx_BASE + CCMR2_OFFSET);

            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // See: P.560 - TIM1&TIM8 capture/compare mode register 1 (TIMx_CCMRx) - OCxM: Output Compare x mode             //
            ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            if (channel == Channel.CH1)
            {
                CCMR1.SetBits(1 << 3); //   <--- Bit 3 OC1PE: Output compare 4 preload enable
                CCMR1.SetBits((uint)mode << 4); //   <--- Bits 6:4 OC1M: Output Compare 1 mode
            }
            else if (channel == Channel.CH2)
            {
                CCMR1.SetBits(1 << 11); //   <--- Bit 11 OC2PE: Output compare 4 preload enable
                CCMR1.SetBits((uint)mode << 12); //   <--- Bits 14:12 OC2M: Output Compare 1 mode
            }
            else if (channel == Channel.CH3)
            {
                CCMR2.SetBits(1 << 3); //   <--- Bit 3 OC3PE: Output compare 4 preload enable
                CCMR2.SetBits((uint)mode << 4); //   <--- Bits 6:4 OC3M: Output Compare 1 mode
            }
            else if (channel == Channel.CH4)
            {
                CCMR2.SetBits(1 << 11); //   <--- Bit 11 OC4PE: Output compare 4 preload enable
                CCMR2.SetBits((uint)mode << 12); //   <--- Bits 14:12 OC4M: Output Compare 1 mode
            }
        }

        public void CompareEnable(UInt32 TIMx_BASE, Channel channel)
        {
            UInt32 CCER_OFFSET = 0x20;
            Register CCER = new Register(TIMx_BASE + CCER_OFFSET);

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bits 1:0 CC1S: Capture/Compare 1 selection                                                                             //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // This bit-field defines the direction of the channel (input/output) as well as the used input.
            //      00: CC1 channel is configured as output
            //      01: CC1 channel is configured as input, IC1 is mapped on TI1
            //      10: CC1 channel is configured as input, IC1 is mapped on TI2
            //      11: CC1 channel is configured as input, IC1 is mapped on TRC. This mode is working only if
            // an internal trigger input is selected through TS bit (TIMx_SMCR register)
            // Note: CC1S bits are writable only when the channel is OFF (CC1E = ‘0’ in TIMx_CCER).
            if (channel == Channel.CH1)
            {
                CCER.ClearBits(0xF << 0);
                CCER.SetBits(1 << 0);  //   <--- Bit 8 CC3E: Capture/Compare 3 output enable
            }
            else if (channel == Channel.CH2)
            {
                CCER.ClearBits(0xF << 4);
                CCER.SetBits(1 << 4);  //   <--- Bit 8 CC3E: Capture/Compare 3 output enable
            }
            else if (channel == Channel.CH3)
            {
                CCER.ClearBits(0xF << 8);
                CCER.SetBits(1 << 8);  //   <--- Bit 8 CC3E: Capture/Compare 3 output enable
            }
            else if (channel == Channel.CH4)
            {
                CCER.ClearBits(0xF << 12);
                CCER.SetBits(1 << 12);  //   <--- Bit 8 CC3E: Capture/Compare 3 output enable
            }
        }

        public void InvertCPUPin(UInt32 TIMx_BASE, Channel channel)
        {
            UInt32 CCER_OFFSET = 0x20;
            Register CCER = new Register(TIMx_BASE + CCER_OFFSET);

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // See: P.565 - capture/compare enable register (TIMx_CCER)                                                               //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bit 1 CC1P: Capture/Compare 1 output polarity
            // CC1 channel configured as output:
            // 0: OC1 active high
            // 1: OC1 active low
            // CC1 channel configured as input:
            // CC1NP/CC1P bits select the active polarity of TI1FP1 and TI2FP1 for trigger or capture operations.
            // 00: non-inverted/rising edge
            // The circuit is sensitive to TIxFP1 rising edge (capture or trigger operations in reset, external
            // clock or trigger mode), TIxFP1 is not inverted (trigger operation in gated mode or encoder mode).
            // 01: inverted/falling edge
            // The circuit is sensitive to TIxFP1 falling edge (capture or trigger operations in reset, external
            // clock or trigger mode), TIxFP1 is inverted (trigger operation in gated mode or encoder mode).
            // 10: reserved, do not use this configuration.
            // 11: non-inverted/both edges

            if (channel == Channel.CH1)
            {
                CCER.ToggleBits(1 << 1);  //   <--- Bit 1 CC1P: Capture/Compare 1 output polarity
            }
            else if (channel == Channel.CH2)
            {
                CCER.ToggleBits(1 << 5);  //   <--- Bit 5 CC2P: Capture/Compare 2 output polarity
            }
            else if (channel == Channel.CH3)
            {
                CCER.ToggleBits(1 << 9);  //   <--- Bit 9 CC3P: Capture/Compare 3 output polarity
            }
            else if (channel == Channel.CH4)
            {
                CCER.ToggleBits(1 << 13);  //   <--- Bit 13 CC4P: Capture/Compare 4 output polarity
            }
        }

        public void MainOutPutEnable(UInt32 TIMx_BASE, bool enable)
        {
            UInt32 BDTR_OFFSET = 0x44;
            Register BDTR = new Register(TIMx_BASE + BDTR_OFFSET);

            if (enable)
            {
                BDTR.SetBits(1 << 15); //   <--- Bit 15 MOE: Main output enable
            }
            else
            {
                BDTR.ClearBits(1 << 15); //   <--- Bit 15 MOE: Main output enable
            }
        }

        // See: P.606 - Using one timer as prescaler for another
        // See: P.607 - Figure 175. Gating timer 2 with OC1REF of timer 1
        // See: P.554 - Table 93. TIMx Internal trigger connection
        // See: P.616 - Table 97. TIMx internal trigger connection
        // See: P.658 - Table 100. TIMx internal trigger connection
        public void MasterModeSelection(CPU.Pin pin, double frequency_Hz, double dutyCycle, CPU.MasterModeSelection masterModeSelection, CPU.UpdateGeneration updateGeneration)
        {
            try
            {
                // Configure current Pin...
                SetPinPort(pin);
            }
            catch
            {
                Debug.Print("Error setting the Pin Info in MasterModeSelection...");
            }


            // Configure Values...
            this.Frequency = frequency_Hz;
            this.Period = PeriodFromFrequency(frequency_Hz, out m_scale);
            this.Duration = DurationFromDutyCycleAndPeriod(dutyCycle, (double)m_period);

            // See: P.606 - Using one timer to enable another timer
            UInt32 CR1_OFFSET = 0x00;
            Register CR1 = new Register(this.TIMx_BASE + CR1_OFFSET);

            UInt32 CR2_OFFSET = 0x04;
            Register CR2 = new Register(this.TIMx_BASE + CR2_OFFSET);

            UInt32 CNT_OFFSET = 0x24;
            Register CNT = new Register(this.TIMx_BASE + CNT_OFFSET);

            UInt32 EGR_OFFSET = 0x14;
            Register EGR = new Register(this.TIMx_BASE + EGR_OFFSET);

            UInt32 SMCR_OFFSET = 0x08;
            Register SMCR = new Register(this.TIMx_BASE + SMCR_OFFSET);

            #region *** Timer synchronization...

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bits 6:4 MMS[2:0]: Master mode selection                                                                               //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            CR2.ClearBits(7 << 4); //   <--- Bits 6:4 MMS[2:0]: Master mode selection
            CR2.SetBits((uint)masterModeSelection << 4); //   <--- Bits 6:4 MMS[2:0]: Master mode selection

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bit 0 UG: Update generation                                                                                            //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            EGR.ClearBits(1 << 0);
            EGR.SetBits((uint)updateGeneration << 0);

            #endregion
        }

        public void SlaveModeSelection(CPU.Pin pin, CPU.SlaveModeSelection mode, CPU.TriggerSelection triggerSelection, CPU.UpdateGeneration updateGeneration, int degreeOfOffSet)
        {
            try
            {
                // Configure current Pin...
                SetPinPort(pin);
            }
            catch
            {
                Debug.Print("Error setting the Pin Info in SlaveModeSelection...");
            }

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // See: P.602 - 18.3.14 Timers and external trigger synchronization                                                       //
            // See: P.606 - Using one timer to enable another timer                                                                   //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

            UInt32 CR1_OFFSET = 0x00;
            Register CR1 = new Register(this.TIMx_BASE + CR1_OFFSET);

            UInt32 CR2_OFFSET = 0x04;
            Register CR2 = new Register(this.TIMx_BASE + CR2_OFFSET);

            UInt32 CNT_OFFSET = 0x24;
            Register CNT = new Register(this.TIMx_BASE + CNT_OFFSET);

            UInt32 EGR_OFFSET = 0x14;
            Register EGR = new Register(this.TIMx_BASE + EGR_OFFSET);

            UInt32 SMCR_OFFSET = 0x08;
            Register SMCR = new Register(this.TIMx_BASE + SMCR_OFFSET);


            #region Slave Mode Selection with Syncronisation...

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bits 2:0 SMS: Slave mode selection                                                                                     //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            SMCR.ClearBits(7 << 0); //   <--- Bits 2:0 SMS: Slave mode selection
            SMCR.SetBits((uint)mode << 0); //   <--- Bits 2:0 SMS: Slave mode selection

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bits 6:4 TS[2:0]: Trigger selection                                                                                    //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // See Table 93: TIMx Internal trigger connection on page 554 for more details on ITRx meaning for each Timer.
            SMCR.ClearBits(7 << 4); //   <--- Bits 6:4 TS: Trigger selection
            SMCR.SetBits((uint)triggerSelection << 4); //   <--- Bits 6:4 TS: Trigger selection

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bit 0 UG: Update generation                                                                                            //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            EGR.ClearBits(1 << 0);
            EGR.SetBits((uint)updateGeneration << 0);

            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // 18.4.10 TIMx counter (TIMx_CNT)                                                                                        //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bits 15:0 CNT[15:0]: Counter value
            CNT.Write((uint)(CycleOffSet(TIMx_BASE, degreeOfOffSet)));

            #endregion
        }

        public void SetSlaveOffSet(CPU.Pin pin, int degreeOfOffSet)
        {
            try
            {
                // Configure current Pin...
                SetPinPort(pin);
            }
            catch
            {
                Debug.Print("Error setting the Pin Info in SlaveModeSelection...");
            }

            UInt32 CNT_OFFSET = 0x24;
            Register CNT = new Register(this.TIMx_BASE + CNT_OFFSET);
            
            #region Slave Mode Selection with Syncronisation...
            
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // 18.4.10 TIMx counter (TIMx_CNT)                                                                                        //
            ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
            // Bits 15:0 CNT[15:0]: Counter value
            CNT.Write((uint)(CycleOffSet(TIMx_BASE, degreeOfOffSet)));

            #endregion
        }

        #endregion

        #region Debug...

        /// <summary>
        /// Debug.Print the Register Name and the current Address...
        /// </summary>
        /// <param name="Name">String Register Display Name: 'CNT' </param>
        /// <param name="Address">The Register to be read: 'CNT'</param>
        private void Print(String Name, Register Register)
        {
            Debug.Print(Name + ": " + Register.Read());
        }

        #endregion
    }

    public static class CPU
    {
        public enum Pin : int
        {
            ///////////////////////////////////////////////////////////////////////////
            // Cerberus PWM uses Timer 1, 3, 8, and 14.                              //
            // Pins are X3, Pin 7, 8 and 9. X4, Pin 7, 8 and 9.                      //
            ///////////////////////////////////////////////////////////////////////////

            Pin7_X3 = 1,
            Pin8_X3 = 2,
            Pin9_X3 = 3,
            Pin7_X4 = 4,
            Pin8_X4 = 5,
            Pin9_X4 = 6
        }

        public enum Clock : uint
        {
            APB1 = 84000000,
            APB2 = 168000000,
        }

        public enum TriggerMode : int
        {
            // It is not allowed to switch from edge-aligned mode to center-aligned mode as long as
            // the counter is enabled (CEN=1)

            EdgeAlignedMode = 0,
            // The counter counts up and down alternatively. Output compare 
            // interrupt flags of channels configured in output (CCxS=00 in TIMx_CCMRx register) 
            // are set only when the counter is counting down.
            CenterAlignedMode1 = 1,
            // The counter counts up and down alternatively. Output compare
            // interrupt flags of channels configured in output (CCxS=00 in TIMx_CCMRx register) are set
            // only when the counter is counting up.
            CenterAlignedMode2 = 2,
            CenterAlignedMode3 = 3
        }

        public enum OCxPort : int
        {
            OC1 = 1,
            OC2 = 2,
            OC3 = 3,
            OC4 = 4
        }

        public enum OCxMode : uint
        {
            Frozen = 0,
            ActiveLevelOnMatch = 1,
            InactiveLevelOnMatch = 2,
            Toggle = 3,
            ForceInactiveLevel = 4,
            ForceActiveLevel = 5,
            PWMMode1 = 6,
            PWMMode2 = 7
        }

        public enum ComparePort : int
        {
            CC1 = 1,
            CC2 = 2,
            CC3 = 3,
            CC4 = 4
        }


        #region TIMx slave mode control register (TIMx_SMCR)

        public enum ExternalTriggerPolarity
        {
        }

        public enum ExternalClockEnable
        {
        }

        public enum ExternalTriggerPrescaler
        {
        }

        public enum ExternalTriggerFilter
        {
        }

        public enum UpdateGeneration
        {
            NoAction = 0,
            DelayedTRGO = 1
        }

        public enum MasterModeSelection
        {
            Reset = 0,
            Enable = 1,
            Update = 2,
            ComparePulse = 3,
            CompareOC1REF = 4,
            CompareOC2REF = 5,
            CompareOC3REF = 6,
            CompareOC4REF = 7,
        }

        public enum TriggerSelection
        {
            ITR0 = 0,
            ITR1 = 1,
            ITR2 = 2,
            ITR3 = 3,
            TI1F_ED = 4,
            TI1FP1 = 5,
            TI2FP2 = 6,
            ETRF = 7
        }

        public enum SlaveModeSelection
        {
            SlaveModeDisabled = 0,
            //EncoderMode1 = 1, < -- Reserved on some Timers....
            //EncoderMode2 = 2, < -- Reserved on some Timers....
            //EncoderMode3 = 3, < -- Reserved on some Timers....
            ResetMode = 4,
            GatedMode = 5,
            TriggerMode = 6,
            ExternalClockMode1 = 7
        }

        #endregion

    }

    public static class OffSet
    {
        public enum Degrees : int
        {
            Deg_0 = 0,
            Deg_1 = 1,
            Deg_2 = 2,
            Deg_3 = 3,
            Deg_4 = 4,
            Deg_5 = 5,
            Deg_6 = 6,
            Deg_7 = 7,
            Deg_8 = 8,
            Deg_9 = 9,
            Deg_10 = 10,
            Deg_11 = 11,
            Deg_12 = 12,
            Deg_13 = 13,
            Deg_14 = 14,
            Deg_15 = 15,
            Deg_16 = 16,
            Deg_17 = 17,
            Deg_18 = 18,
            Deg_19 = 19,
            Deg_20 = 20,
            Deg_21 = 21,
            Deg_22 = 22,
            Deg_23 = 23,
            Deg_24 = 24,
            Deg_25 = 25,
            Deg_26 = 26,
            Deg_27 = 27,
            Deg_28 = 28,
            Deg_29 = 29,
            Deg_30 = 30,
            Deg_31 = 31,
            Deg_32 = 32,
            Deg_33 = 33,
            Deg_34 = 34,
            Deg_35 = 35,
            Deg_36 = 36,
            Deg_37 = 37,
            Deg_38 = 38,
            Deg_39 = 39,
            Deg_40 = 40,
            Deg_41 = 41,
            Deg_42 = 42,
            Deg_43 = 43,
            Deg_44 = 44,
            Deg_45 = 45,
            Deg_46 = 46,
            Deg_47 = 47,
            Deg_48 = 48,
            Deg_49 = 49,
            Deg_50 = 50,
            Deg_51 = 51,
            Deg_52 = 52,
            Deg_53 = 53,
            Deg_54 = 54,
            Deg_55 = 55,
            Deg_56 = 56,
            Deg_57 = 57,
            Deg_58 = 58,
            Deg_59 = 59,
            Deg_60 = 60,
            Deg_61 = 61,
            Deg_62 = 62,
            Deg_63 = 63,
            Deg_64 = 64,
            Deg_65 = 65,
            Deg_66 = 66,
            Deg_67 = 67,
            Deg_68 = 68,
            Deg_69 = 69,
            Deg_70 = 70,
            Deg_71 = 71,
            Deg_72 = 72,
            Deg_73 = 73,
            Deg_74 = 74,
            Deg_75 = 75,
            Deg_76 = 76,
            Deg_77 = 77,
            Deg_78 = 78,
            Deg_79 = 79,
            Deg_80 = 80,
            Deg_81 = 81,
            Deg_82 = 82,
            Deg_83 = 83,
            Deg_84 = 84,
            Deg_85 = 85,
            Deg_86 = 86,
            Deg_87 = 87,
            Deg_88 = 88,
            Deg_89 = 89,
            Deg_90 = 90,
            Deg_91 = 91,
            Deg_92 = 92,
            Deg_93 = 93,
            Deg_94 = 94,
            Deg_95 = 95,
            Deg_96 = 96,
            Deg_97 = 97,
            Deg_98 = 98,
            Deg_99 = 99,
            Deg_100 = 100,
            Deg_101 = 101,
            Deg_102 = 102,
            Deg_103 = 103,
            Deg_104 = 104,
            Deg_105 = 105,
            Deg_106 = 106,
            Deg_107 = 107,
            Deg_108 = 108,
            Deg_109 = 109,
            Deg_110 = 110,
            Deg_111 = 111,
            Deg_112 = 112,
            Deg_113 = 113,
            Deg_114 = 114,
            Deg_115 = 115,
            Deg_116 = 116,
            Deg_117 = 117,
            Deg_118 = 118,
            Deg_119 = 119,
            Deg_120 = 120,
            Deg_121 = 121,
            Deg_122 = 122,
            Deg_123 = 123,
            Deg_124 = 124,
            Deg_125 = 125,
            Deg_126 = 126,
            Deg_127 = 127,
            Deg_128 = 128,
            Deg_129 = 129,
            Deg_130 = 130,
            Deg_131 = 131,
            Deg_132 = 132,
            Deg_133 = 133,
            Deg_134 = 134,
            Deg_135 = 135,
            Deg_136 = 136,
            Deg_137 = 137,
            Deg_138 = 138,
            Deg_139 = 139,
            Deg_140 = 140,
            Deg_141 = 141,
            Deg_142 = 142,
            Deg_143 = 143,
            Deg_144 = 144,
            Deg_145 = 145,
            Deg_146 = 146,
            Deg_147 = 147,
            Deg_148 = 148,
            Deg_149 = 149,
            Deg_150 = 150,
            Deg_151 = 151,
            Deg_152 = 152,
            Deg_153 = 153,
            Deg_154 = 154,
            Deg_155 = 155,
            Deg_156 = 156,
            Deg_157 = 157,
            Deg_158 = 158,
            Deg_159 = 159,
            Deg_160 = 160,
            Deg_161 = 161,
            Deg_162 = 162,
            Deg_163 = 163,
            Deg_164 = 164,
            Deg_165 = 165,
            Deg_166 = 166,
            Deg_167 = 167,
            Deg_168 = 168,
            Deg_169 = 169,
            Deg_170 = 170,
            Deg_171 = 171,
            Deg_172 = 172,
            Deg_173 = 173,
            Deg_174 = 174,
            Deg_175 = 175,
            Deg_176 = 176,
            Deg_177 = 177,
            Deg_178 = 178,
            Deg_179 = 179,
            Deg_180 = 180,
            Deg_181 = 181,
            Deg_182 = 182,
            Deg_183 = 183,
            Deg_184 = 184,
            Deg_185 = 185,
            Deg_186 = 186,
            Deg_187 = 187,
            Deg_188 = 188,
            Deg_189 = 189,
            Deg_190 = 190,
            Deg_191 = 191,
            Deg_192 = 192,
            Deg_193 = 193,
            Deg_194 = 194,
            Deg_195 = 195,
            Deg_196 = 196,
            Deg_197 = 197,
            Deg_198 = 198,
            Deg_199 = 199,
            Deg_200 = 200,
            Deg_201 = 201,
            Deg_202 = 202,
            Deg_203 = 203,
            Deg_204 = 204,
            Deg_205 = 205,
            Deg_206 = 206,
            Deg_207 = 207,
            Deg_208 = 208,
            Deg_209 = 209,
            Deg_210 = 210,
            Deg_211 = 211,
            Deg_212 = 212,
            Deg_213 = 213,
            Deg_214 = 214,
            Deg_215 = 215,
            Deg_216 = 216,
            Deg_217 = 217,
            Deg_218 = 218,
            Deg_219 = 219,
            Deg_220 = 220,
            Deg_221 = 221,
            Deg_222 = 222,
            Deg_223 = 223,
            Deg_224 = 224,
            Deg_225 = 225,
            Deg_226 = 226,
            Deg_227 = 227,
            Deg_228 = 228,
            Deg_229 = 229,
            Deg_230 = 230,
            Deg_231 = 231,
            Deg_232 = 232,
            Deg_233 = 233,
            Deg_234 = 234,
            Deg_235 = 235,
            Deg_236 = 236,
            Deg_237 = 237,
            Deg_238 = 238,
            Deg_239 = 239,
            Deg_240 = 240,
            Deg_241 = 241,
            Deg_242 = 242,
            Deg_243 = 243,
            Deg_244 = 244,
            Deg_245 = 245,
            Deg_246 = 246,
            Deg_247 = 247,
            Deg_248 = 248,
            Deg_249 = 249,
            Deg_250 = 250,
            Deg_251 = 251,
            Deg_252 = 252,
            Deg_253 = 253,
            Deg_254 = 254,
            Deg_255 = 255,
            Deg_256 = 256,
            Deg_257 = 257,
            Deg_258 = 258,
            Deg_259 = 259,
            Deg_260 = 260,
            Deg_261 = 261,
            Deg_262 = 262,
            Deg_263 = 263,
            Deg_264 = 264,
            Deg_265 = 265,
            Deg_266 = 266,
            Deg_267 = 267,
            Deg_268 = 268,
            Deg_269 = 269,
            Deg_270 = 270,
            Deg_271 = 271,
            Deg_272 = 272,
            Deg_273 = 273,
            Deg_274 = 274,
            Deg_275 = 275,
            Deg_276 = 276,
            Deg_277 = 277,
            Deg_278 = 278,
            Deg_279 = 279,
            Deg_280 = 280,
            Deg_281 = 281,
            Deg_282 = 282,
            Deg_283 = 283,
            Deg_284 = 284,
            Deg_285 = 285,
            Deg_286 = 286,
            Deg_287 = 287,
            Deg_288 = 288,
            Deg_289 = 289,
            Deg_290 = 290,
            Deg_291 = 291,
            Deg_292 = 292,
            Deg_293 = 293,
            Deg_294 = 294,
            Deg_295 = 295,
            Deg_296 = 296,
            Deg_297 = 297,
            Deg_298 = 298,
            Deg_299 = 299,
            Deg_300 = 300,
            Deg_301 = 301,
            Deg_302 = 302,
            Deg_303 = 303,
            Deg_304 = 304,
            Deg_305 = 305,
            Deg_306 = 306,
            Deg_307 = 307,
            Deg_308 = 308,
            Deg_309 = 309,
            Deg_310 = 310,
            Deg_311 = 311,
            Deg_312 = 312,
            Deg_313 = 313,
            Deg_314 = 314,
            Deg_315 = 315,
            Deg_316 = 316,
            Deg_317 = 317,
            Deg_318 = 318,
            Deg_319 = 319,
            Deg_320 = 320,
            Deg_321 = 321,
            Deg_322 = 322,
            Deg_323 = 323,
            Deg_324 = 324,
            Deg_325 = 325,
            Deg_326 = 326,
            Deg_327 = 327,
            Deg_328 = 328,
            Deg_329 = 329,
            Deg_330 = 330,
            Deg_331 = 331,
            Deg_332 = 332,
            Deg_333 = 333,
            Deg_334 = 334,
            Deg_335 = 335,
            Deg_336 = 336,
            Deg_337 = 337,
            Deg_338 = 338,
            Deg_339 = 339,
            Deg_340 = 340,
            Deg_341 = 341,
            Deg_342 = 342,
            Deg_343 = 343,
            Deg_344 = 344,
            Deg_345 = 345,
            Deg_346 = 346,
            Deg_347 = 347,
            Deg_348 = 348,
            Deg_349 = 349,
            Deg_350 = 350,
            Deg_351 = 351,
            Deg_352 = 352,
            Deg_353 = 353,
            Deg_354 = 354,
            Deg_355 = 355,
            Deg_356 = 356,
            Deg_357 = 357,
            Deg_358 = 358,
            Deg_359 = 359,
            Deg_360 = 360
        }
    }
}
